# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/optimizer.foreach.ipynb.

# %% ../../nbs/optimizer.foreach.ipynb 3
from __future__ import annotations

import numpy as np

from fastai.optimizer import Optimizer

from ..imports import *

# %% auto 0
__all__ = []

# %% ../../nbs/optimizer.foreach.ipynb 8
class ForEachOptimizer(Optimizer):
    "Base foreach optimizer class, updating `params` with `opt_step` instead of `Optimizer.cbs`"
    def __init__(self,
        params:Listified[Tensor], # Model parameters
        opt_step:Callable, # `ForEachOptimizer` optimizer step
        decouple_wd:bool=True, # Use true weight decay or L2 regularization, if applicable
        **defaults # Optimizer specific hyper parameters
    ):
        if notmax_torch('1.12'):
            warn(f'ForEach optimizers are untested on PyTorch {torch.__verson__}, recommended to use 1.12 or newer')
        super().__init__(params, [None], True, **defaults)
        self.opt_step = opt_step
        self.decouple_wd = decouple_wd

# %% ../../nbs/optimizer.foreach.ipynb 11
def sgd_foreach_step(p:list[Tensor], g:list[Tensor], grad_avg:list[Tensor|None], ones:list[Tensor|None], 
                     do_wd:np.ndarray[Any, bool], lr:float, wd:float, mom:float, decouple_wd:bool, **kwargs):

    if wd != 0:
        if decouple_wd:
            # weight_decay
            wd = np.where(do_wd, 1-lr*wd, 1.)
            torch._foreach_mul_(p, scalars=wd.tolist())
        else:
            # l2_reg
            wd = np.where(do_wd, wd, 1.)
            torch._foreach_addcdiv_(g, p, ones, scalars=wd.tolist())
            # cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv

    if mom != 0:
        # average_grad
        torch._foreach_mul_(grad_avg, mom)
        torch._foreach_add_(grad_avg, g)

        # momentum_step
        torch._foreach_add_(p, grad_avg, alpha=-lr)
    else:
        # sgd_step
        torch._foreach_add_(p, g, alpha=-lr)

# %% ../../nbs/optimizer.foreach.ipynb 12
class SGDForEachOptimizer(ForEachOptimizer):
    "A `ForEachOptimizer` with a modified step for `sgd_foreach_step`"
    @torch.no_grad()
    def step(self, closure=None):
        if closure is not None:
            raise NotImplementedError("fastai optimizers currently do not support closure")
        for pg, hyper in zip(self.param_lists, self.hypers):
            pl, gl, grad_avg, ones, do_wd = [], [], [], [], []

            for p in pg:
                if hasattr(p, 'grad') and p.grad is not None:
                    state = self.state[p]

                    if 'setup' not in state:
                        if hyper['mom'] != 0:
                            state['grad_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        if not self.decouple_wd:
                            state['ones'] = torch.ones_like(p, memory_format=torch.preserve_format)
                        state['setup'] = True

                    pl.append(p)
                    gl.append(p.grad)
                    grad_avg.append(state.get('grad_avg', None))
                    ones.append(state.get('ones', None))
                    do_wd.append(state.get('do_wd', True))

            self.opt_step(p=pl, g=gl, grad_avg=grad_avg, ones=ones, do_wd=np.array(do_wd, dtype=bool),
                          decouple_wd=self.decouple_wd, **hyper)

# %% ../../nbs/optimizer.foreach.ipynb 20
def adam_foreach_step(p:list[Tensor], g:list[Tensor], grad_avg:list[Tensor], sqr_avg:list[Tensor], ones:list[Tensor|None],
                      steps:np.ndarray[Any, int], do_wd:np.ndarray[Any, bool], lr:float, wd:float, mom:float, sqr_mom:float,
                      eps:float, decouple_wd:bool, **kwargs):

    if wd != 0:
        if decouple_wd:
            # weight_decay
            wd = np.where(do_wd, 1-lr*wd, 1.)
            torch._foreach_mul_(p, scalars=wd.tolist())
        else:
            # l2_reg
            wd = np.where(do_wd, wd, 1.)
            torch._foreach_addcdiv_(g, p, ones, scalars=wd.tolist())
            # cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv

    # average_grad, dampening=True
    torch._foreach_mul_(grad_avg, mom)
    torch._foreach_add_(grad_avg, g, alpha=1-mom)

    # average_sqr_grad
    torch._foreach_mul_(sqr_avg, sqr_mom)
    torch._foreach_addcmul_(sqr_avg, g, g, value=1-sqr_mom)

    # adam_step
    debias1 = -lr / (1 - mom**steps)
    debias2 = np.sqrt(1 - sqr_mom**steps)

    sqr_avg_debias2 = torch._foreach_sqrt(sqr_avg)
    torch._foreach_div_(sqr_avg_debias2, debias2.tolist())
    torch._foreach_add_(sqr_avg_debias2, eps)

    torch._foreach_addcdiv_(p, grad_avg, sqr_avg_debias2, debias1.tolist())

# %% ../../nbs/optimizer.foreach.ipynb 21
class AdamForEachOptimizer(ForEachOptimizer):
    "An `ForEachOptimizer` with a modified step for `adam_foreach_step`"
    @torch.no_grad()
    def step(self, closure=None):
        if closure is not None:
            raise NotImplementedError("fastai optimizers currently do not support closure")
        for pg, hyper in zip(self.param_lists, self.hypers):
            pl, gl, grad_avg, sqr_avg, ones, steps, do_wd = [], [], [], [], [], [], []

            for p in pg:
                if hasattr(p, 'grad') and p.grad is not None:
                    state = self.state[p]

                    if 'step' not in state:
                        state['grad_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['sqr_avg']  = torch.zeros_like(p, memory_format=torch.preserve_format)
                        if not self.decouple_wd:
                            state['ones'] = torch.ones_like(p, memory_format=torch.preserve_format)
                        state['step'] = 0

                    state['step'] += 1
                    pl.append(p)
                    gl.append(p.grad)
                    grad_avg.append(state['grad_avg'])
                    sqr_avg.append(state['sqr_avg'])
                    ones.append(state.get('ones', None))
                    steps.append(state['step'])
                    do_wd.append(state.get('do_wd', True))

            self.opt_step(p=pl, g=gl, grad_avg=grad_avg, sqr_avg=sqr_avg, ones=ones,
                          steps=np.array(steps, dtype=np.int32), do_wd=np.array(do_wd, dtype=bool),
                          decouple_wd=self.decouple_wd, **hyper)

# %% ../../nbs/optimizer.foreach.ipynb 28
def radam_foreach_step(p:list[Tensor], g:list[Tensor], grad_avg:list[Tensor], sqr_avg:list[Tensor], ones:list[Tensor],
                       steps:np.ndarray[Any, int], do_wd:np.ndarray[Any, bool], lr:float, wd:float, mom:float, sqr_mom:float,
                       eps:float, decouple_wd:bool, **kwargs):

    if wd != 0:
        if decouple_wd:
            # weight_decay
            wd = np.where(do_wd, 1-lr*wd, 1.)
            torch._foreach_mul_(p, scalars=wd.tolist())
        else:
            # l2_reg
            wd = np.where(do_wd, wd, 1.)
            torch._foreach_addcdiv_(g, p, ones, scalars=wd.tolist())
            # cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv

    # average_grad, dampening=True
    torch._foreach_mul_(grad_avg, mom)
    torch._foreach_add_(grad_avg, g, alpha=1-mom)

    # average_sqr_grad
    torch._foreach_mul_(sqr_avg, sqr_mom)
    torch._foreach_addcmul_(sqr_avg, g, g, value=1-sqr_mom)

    # radam_step
    debias1 = -lr / (1 - mom**steps)
    debias2 = np.sqrt(1 - sqr_mom**steps)

    r_inf = 2/(1-sqr_mom) - 1
    r = r_inf - 2*steps*sqr_mom**steps/(1-sqr_mom**steps)

    rect   = np.where(r>5, debias1*np.sqrt(((r-4) * (r-2) * r_inf)/((r_inf-4)*(r_inf-2)*r), where=r>5), 0)
    unrect = np.where(r<=5, debias1, 0)

    # rectified step
    sqrt_avg_debias2 = torch._foreach_sqrt(sqr_avg)
    torch._foreach_div_(sqrt_avg_debias2, debias2.tolist())
    torch._foreach_add_(sqrt_avg_debias2, eps)
    torch._foreach_addcdiv_(p, grad_avg, sqrt_avg_debias2, scalars=rect.tolist())

    # unrectified step. cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv
    torch._foreach_addcdiv_(p, grad_avg, ones, scalars=unrect.tolist())

# %% ../../nbs/optimizer.foreach.ipynb 29
class RAdamForEachOptimizer(ForEachOptimizer):
    "An `ForEachOptimizer` with a modified step for `radam_foreach_step`"
    @torch.no_grad()
    def step(self, closure=None):
        if closure is not None:
            raise NotImplementedError("fastai optimizers currently do not support closure")
        for pg, hyper in zip(self.param_lists, self.hypers):
            pl, gl, grad_avg, sqr_avg, steps, ones, do_wd = [], [], [], [], [], [], []

            for p in pg:
                if hasattr(p, 'grad') and p.grad is not None:
                    state = self.state[p]

                    if 'step' not in state:
                        state['grad_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['sqr_avg']  = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['ones']     = torch.ones_like(p, memory_format=torch.preserve_format)
                        state['step']     = 0

                    state['step'] += 1
                    pl.append(p)
                    gl.append(p.grad)
                    grad_avg.append(state['grad_avg'])
                    sqr_avg.append(state['sqr_avg'])
                    ones.append(state['ones'])
                    steps.append(state['step'])
                    do_wd.append(state.get('do_wd', True))

            self.opt_step(p=pl, g=gl, grad_avg=grad_avg, sqr_avg=sqr_avg, ones=ones,
                          steps=np.array(steps, dtype=np.int32), do_wd=np.array(do_wd, dtype=bool),
                          decouple_wd=self.decouple_wd, **hyper)

# %% ../../nbs/optimizer.foreach.ipynb 35
@torch.jit.script
def lamb_jit_substep(p:Tensor, lstep:Tensor, lr:float):
    r1 = p.pow(2).mean().sqrt()
    r2 = lstep.pow(2).mean().sqrt()
    if r1 == 0 or r2 == 0:
        return -lr
    else:
        return -lr*min(r1/r2, 10.)

# %% ../../nbs/optimizer.foreach.ipynb 36
def lamb_foreach_step(p:list[Tensor], g:list[Tensor], grad_avg:list[Tensor], sqr_avg:list[Tensor], ones:list[Tensor],
                      steps:np.ndarray[Any, int], do_wd:np.ndarray[Any, bool], lr:float, wd:float, mom:float, sqr_mom:float,
                      eps:float, decouple_wd:bool, **kwargs):

    if wd != 0:
        if decouple_wd:
            # weight_decay
            wd = np.where(do_wd, 1-lr*wd, 1.)
            torch._foreach_mul_(p, scalars=wd.tolist())
        else:
            # l2_reg
            wd = np.where(do_wd, wd, 1.)
            torch._foreach_addcdiv_(g, p, ones, scalars=wd.tolist())
            # cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv

    # average_grad, dampening=True
    torch._foreach_mul_(grad_avg, mom)
    torch._foreach_add_(grad_avg, g, alpha=1-mom)

    # average_sqr_grad
    torch._foreach_mul_(sqr_avg, sqr_mom)
    torch._foreach_addcmul_(sqr_avg, g, g, value=1-sqr_mom)

    # lamb_step
    debias1 = 1 - mom**steps
    debias2 = np.sqrt(1 - sqr_mom**steps)

    debias2 = torch._foreach_div(torch._foreach_sqrt(sqr_avg), debias2.tolist())
    torch._foreach_add_(debias2, eps)
    lstep = torch._foreach_div(grad_avg, debias1.tolist())
    torch._foreach_div_(lstep, debias2)

    # there are no implementations for foreach_pow, foreach_mean, or foreach_where/if methods
    q = [lamb_jit_substep(pi, ls, lr) for pi, ls in zip(p, lstep)]

    # cannot use scalers with foreach_add & multiple tensors, so divide by one with foreach_addcdiv
    torch._foreach_addcdiv_(p, lstep, ones, scalars=q)

# %% ../../nbs/optimizer.foreach.ipynb 37
class LambForEachOptimizer(RAdamForEachOptimizer):
    "An `ForEachOptimizer` with a modified step for `lamb_foreach_step`"

# %% ../../nbs/optimizer.foreach.ipynb 43
def ranger_foreach_step(p:list[Tensor], g:list[Tensor], grad_avg:list[Tensor], sqr_avg:list[Tensor], slow_p:list[Tensor],
                        ones:list[Tensor], steps:np.ndarray[Any, int], do_wd:np.ndarray[Any, bool], lr:float, wd:float,
                        mom:float, sqr_mom:float, eps:float, decouple_wd:bool, count:int, k:int, alpha:float, **kwargs):

    radam_foreach_step(p=p, g=g, grad_avg=grad_avg, sqr_avg=sqr_avg, ones=ones, steps=steps, do_wd=do_wd,
                       lr=lr, wd=wd, mom=mom, sqr_mom=sqr_mom, eps=eps, decouple_wd=decouple_wd)

    if count % k == 0:
        torch._foreach_add_(slow_p, torch._foreach_sub(p, slow_p), alpha=alpha)
        # there currently is no foreach_set method
        [pi.set_(slow_pi.clone()) for pi, slow_pi in zip(p, slow_p)]

# %% ../../nbs/optimizer.foreach.ipynb 44
class RangerForEachOptimizer(ForEachOptimizer):
    "An `ForEachOptimizer` with a modified `LookAhead` step for `ranger_foreach_step`"
    def __init__(self,
        params:Listified[Tensor], # Model parameters
        opt_step:Callable, # `ForEachOptimizer` optimizer step
        decouple_wd:bool=True, # Use true weight decay or L2 regularization, if applicable
        **defaults # Optimizer specific hyper parameters default values
    ):
        super().__init__(params, opt_step, decouple_wd, **defaults)
        self._init_state()

    @torch.no_grad()
    def step(self, closure=None):
        self.count += 1
        if closure is not None:
            raise NotImplementedError("fastai optimizers currently do not support closure")
        for pg, hyper in zip(self.param_lists, self.hypers):
            pl, gl, grad_avg, sqr_avg, slow_p, steps, ones, do_wd = [], [], [], [], [], [], [], []
            for p in pg:
                if hasattr(p, 'grad') and p.grad is not None:
                    state = self.state[p]

                    if 'step' not in state:
                        state['grad_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['sqr_avg']  = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['ones']     = torch.ones_like(p, memory_format=torch.preserve_format)
                        state['slow_p']   = p.data.clone()
                        state['step']     = 0

                    state['step'] += 1
                    pl.append(p)
                    gl.append(p.grad)
                    grad_avg.append(state['grad_avg'])
                    sqr_avg.append(state['sqr_avg'])
                    slow_p.append(state['slow_p'])
                    ones.append(state['ones'])
                    steps.append(state['step'])
                    do_wd.append(state.get('do_wd', True))

            self.opt_step(p=pl, g=gl, grad_avg=grad_avg, sqr_avg=sqr_avg, slow_p=slow_p, ones=ones,
                          steps=np.array(steps, dtype=np.int32), do_wd=np.array(do_wd, dtype=bool),
                          decouple_wd=self.decouple_wd, count=self.count, **hyper)

    def clear_state(self):
        super().clear_state()
        self._init_state()

    def state_dict(self):
        state = super().state_dict()
        state.update({'count': self.count})
        return state

    def load_state_dict(self, sd):
        self.count = sd.pop('count')
        super().load_state_dict(sd)

    def _init_state(self):
        self.count = 0



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/optimizer.adan.ipynb.

# %% ../../nbs/optimizer.adan.ipynb 1
# Memory and operations reduction ported from the official Adan implementation
# https://github.com/sail-sg/Adan - Apache License 2.0 - Copyright 2022 Xingyu Xie et al

# %% ../../nbs/optimizer.adan.ipynb 3
from __future__ import annotations
from typing import Optional, Dict

import numpy as np

from fastai.optimizer import Optimizer

from .foreach import ForEachOptimizer
from .torchscript import JitOptimizer
from ..imports import *

# %% auto 0
__all__ = ['Adan', 'adan', 'AdanLargeBatchLR']

# %% ../../nbs/optimizer.adan.ipynb 5
def debias(beta:float, step:int):
    "Simple debias calculation"
    return 1-beta**step

# %% ../../nbs/optimizer.adan.ipynb 7
def adan_setup(p:Tensor, step:int=0, grad_avg:Tensor|None=None, diff_avg:Tensor|None=None,
               sqr_avg:Tensor|None=None, prior_grad:Tensor|None=None, paper_init:bool=False, **kwargs):
    "Handles Adan setup and keeps track of steps"
    if step == 0:
        grad_avg = torch.zeros_like(p, memory_format=torch.preserve_format)
        diff_avg = torch.zeros_like(p, memory_format=torch.preserve_format)
        sqr_avg  = torch.zeros_like(p, memory_format=torch.preserve_format)
        if paper_init:
            prior_grad = p.grad.clone().mul_(-1)
        else:
            prior_grad = torch.zeros_like(p, memory_format=torch.preserve_format)
        step += 1
        return {'grad_avg':grad_avg, 'diff_avg':diff_avg, 'sqr_avg':sqr_avg, 'prior_grad':prior_grad, 'step':step}
    else:
        step += 1
        return {'step':step}

# %% ../../nbs/optimizer.adan.ipynb 8
def adan_step(p:Tensor, lr:float, eps:float, wd:float, beta1:float, beta2:float, beta3:float,
              step:int, grad_avg:Tensor, diff_avg:Tensor, sqr_avg:Tensor, prior_grad:Tensor,
              do_wd:bool=True, **kwargs):
    "Updates Adan moving averages and performs the Adan step with `lr` on `p`"

    # difference between current and previous gradients, prior_grad is negated in last step
    prior_grad = torch.add(p.grad.data, prior_grad)

    # update m_k
    grad_avg.mul_(beta1).add_(p.grad.data, alpha=1-beta1)

    # update v_k
    diff_avg.mul_(beta2).add_(prior_grad, alpha=1-beta2)

    # update n_k
    prior_grad = torch.add(p.grad.data, prior_grad, alpha=beta2)
    sqr_avg.mul_(beta3).addcmul_(prior_grad, prior_grad, value=1-beta3)

    # calculate debias terms
    db1 = 1/debias(beta1, step)
    db2 = beta2/debias(beta2, step)
    db3 = lr*math.sqrt(debias(beta3, step))

    # calculate applied λ
    wd = (1+lr*wd) if wd!=0 and do_wd else 1

    # calculate η_k
    lrs = torch.sqrt(sqr_avg).div(db3).add(eps)

    # perform Adan step and apply to parameter `p`
    p.data.addcdiv_(grad_avg, lrs, value=-db1)
    p.data.addcdiv_(diff_avg, lrs, value=-db2)
    p.data.div_(wd)

    # set next step's prior_grad as negated current grad
    prior_grad = p.grad.data.clone().mul_(-1)
    return {'grad_avg':grad_avg, 'diff_avg':diff_avg, 'sqr_avg':sqr_avg, 'prior_grad':prior_grad}

adan_step.defaults = dict(beta1=0.98, beta2=0.92, beta3=0.99)

# %% ../../nbs/optimizer.adan.ipynb 10
@torch.jit.script
def adan_jit_step(p:Tensor, g:Tensor, lr:float, wd:float, beta1:float, beta2:float, beta3:float, eps:float,
                  paper_init:bool, grad_avg:Optional[Tensor]=None, diff_avg:Optional[Tensor]=None,
                  sqr_avg:Optional[Tensor]=None, prior_grad:Optional[Tensor]=None, do_wd:bool=True, step:int=0,
                  force_train:Optional[bool]=None, mom:Optional[float]=None, decouple_wd:bool=False):
    dp = p
    grad = g
    step += 1

    if grad_avg is None:
        grad_avg = torch.zeros_like(dp, memory_format=torch.preserve_format)
    if diff_avg is None:
        diff_avg = torch.zeros_like(dp, memory_format=torch.preserve_format)
    if sqr_avg is None:
        sqr_avg = torch.zeros_like(dp, memory_format=torch.preserve_format)
    if prior_grad is None:
        if paper_init:
            prior_grad = grad.clone()
        else:
            prior_grad = torch.zeros_like(dp, memory_format=torch.preserve_format)

    # difference between current and previous gradients
    diff_grad = grad.sub(prior_grad)

    # update m_k
    grad_avg = grad_avg.mul(beta1).add(grad, alpha=1-beta1)

    # update v_k
    diff_avg = diff_avg.mul(beta2).add(diff_grad, alpha=1-beta2)

    # update n_k
    adjusted_grad = grad.add(diff_grad, alpha=beta2)
    sqr_avg = sqr_avg.mul(beta3).addcmul(adjusted_grad, adjusted_grad, value=1-beta3)

    # calculate debias terms
    db1 = 1-beta1**step
    db2 = 1-beta2**step
    db3 = math.sqrt(1-beta3**step)

    # calculate applied λ
    if wd!=0 and do_wd:
        wd = (1+lr*wd)
    else:
        wd = 1.

    # calculate η_k
    lrs = lr/torch.sqrt(sqr_avg).div(db3).add(eps)

    # perform Adan step
    dp = dp.sub(torch.add(grad_avg.div(db1), diff_avg.div(db2), alpha=beta2).mul(lrs)).div(wd)

    # set next step's prior_grad as negated current grad
    prior_grad = grad.clone()

    # apply results to parameter p
    p.set_(dp)
    g.set_(grad)

    return torch.jit.annotate(Dict[str, Union[Tensor, int]], {'grad_avg':grad_avg, 'diff_avg':diff_avg, 'sqr_avg':sqr_avg, 'prior_grad':prior_grad, 'step':step})

# %% ../../nbs/optimizer.adan.ipynb 12
def adan_foreach_step(p:list[Tensor], grad:list[Tensor], grad_avg:list[Tensor], diff_avg:list[Tensor],
                      sqr_avg:list[Tensor], prior_grad:list[Tensor], steps:np.ndarray[Any, int],
                      do_wd:np.ndarray[Any, bool], lr:float, wd:float, beta1:float, beta2:float,
                      beta3:float, eps:float, **kwargs):

    # difference between current and previous gradients, prior_grad is negated in last step
    torch._foreach_add_(prior_grad, grad)

    # update m_k
    torch._foreach_mul_(grad_avg, beta1)
    torch._foreach_add_(grad_avg, grad, alpha=1-beta1)

    # update v_k
    torch._foreach_mul_(diff_avg, beta2)
    torch._foreach_add_(diff_avg, prior_grad, alpha=1-beta2)

    # update n_k
    torch._foreach_mul_(prior_grad, scalar=beta2)
    torch._foreach_add_(prior_grad, grad)
    torch._foreach_mul_(sqr_avg, beta3)
    torch._foreach_addcmul_(sqr_avg, prior_grad, prior_grad, value=1-beta3)

    # calculate debias terms
    db1 = -1/(1 - beta1**steps)
    db2 = -beta2/(1 - beta2**steps)
    db3 = lr*np.sqrt(1 - beta3**steps)

    # calculate η_k
    lrs = torch._foreach_sqrt(sqr_avg)
    torch._foreach_div_(lrs, scalars=db3.tolist())
    torch._foreach_add_(lrs, scalar=eps)

    # perform Adan step
    torch._foreach_addcdiv_(p, grad_avg, lrs, scalars=db1.tolist())
    torch._foreach_addcdiv_(p, diff_avg, lrs, scalars=db2.tolist())

    # calculate and apply λ
    if wd != 0:
        wd = np.where(do_wd, 1+lr*wd, 1.)
        torch._foreach_div_(p, scalars=wd.tolist())

    # set next step's prior_grad as negated current grad
    torch._foreach_zero_(prior_grad)
    torch._foreach_add_(prior_grad, grad, alpha=-1)

# %% ../../nbs/optimizer.adan.ipynb 13
class AdanForEachOptimizer(ForEachOptimizer):
    "An `Optimizer` with a modified step for Adan ForEach"
    def __init__(self,
        params:Listified[Tensor], # Model parameters
        opt_step:Callable, # `ForEachOptimizer` optimizer step
        paper_init:bool=False, # Initialize first prior_grad to grad following paper or zeros
        **defaults # Optimizer specific hyper parameters
    ):
        super().__init__(params, opt_step, **defaults)
        self.paper_init = paper_init

    @torch.no_grad()
    def step(self, closure=None):
        if closure is not None: raise NotImplementedError("fastai optimizers currently do not support closure")
        for pg, hyper in zip(self.param_lists, self.hypers):
            pl, gl, grad_avg, diff_avg, sqr_avg, prior_grad, steps, do_wd = [], [], [], [], [], [], [], []

            for p in pg:
                if hasattr(p, 'grad') and p.grad is not None:
                    state = self.state[p]

                    if 'step' not in state:
                        state['grad_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['diff_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['sqr_avg']  = torch.zeros_like(p, memory_format=torch.preserve_format)
                        if self.paper_init:
                            state['prior_grad'] = p.grad.clone().mul_(-1)
                        else:
                            state['prior_grad'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        state['step'] = 0

                    state['step'] += 1
                    pl.append(p)
                    gl.append(p.grad)
                    grad_avg.append(state['grad_avg'])
                    diff_avg.append(state['diff_avg'])
                    sqr_avg.append(state['sqr_avg'])
                    prior_grad.append(state['prior_grad'])
                    do_wd.append(state.get('do_wd', True))
                    steps.append(state['step'])

            self.opt_step(p=pl, grad=gl, grad_avg=grad_avg, diff_avg=diff_avg, sqr_avg=sqr_avg,
                          prior_grad=prior_grad, steps=np.array(steps, dtype=np.int32), do_wd=np.array(do_wd, dtype=bool), **hyper)

# %% ../../nbs/optimizer.adan.ipynb 16
def Adan(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    beta1:float=0.98, # Gradient moving average (β1) coefficient
    beta2:float=0.92, # Gradient difference moving average (β2) coefficient
    beta3:float=0.99, # Gradient squared moving average (β3) coefficient
    eps:float=1e-8, # Added for numerical stability
    wd:float=0.02, # True weight decay
    paper_init:bool=False, # Initialize prior gradient with current gradient per paper, or zeroes
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|AdanForEachOptimizer|JitOptimizer:
    "A fastai Adan optimizer with optional ForEach and TorchScript implementations"
    if foreach:
        return AdanForEachOptimizer(params, adan_foreach_step, lr=lr, beta1=beta1, beta2=beta2,
                                    beta3=beta3, eps=eps, wd=wd, paper_init=paper_init)
    elif jit:
        cb = partial(adan_jit_step, paper_init=paper_init)
        return JitOptimizer(params, cb, lr=lr, beta1=beta1, beta2=beta2, beta3=beta3, eps=eps, wd=wd)
    else:
        cbs = [partial(adan_setup, paper_init=paper_init), adan_step]
        return Optimizer(params, cbs, lr=lr, beta1=beta1, beta2=beta2, beta3=beta3, eps=eps, wd=wd)

# %% ../../nbs/optimizer.adan.ipynb 17
def adan(
    beta1:float=0.98, # Gradient moving average (β1) coefficient
    beta2:float=0.92, # Gradient difference moving average (β2) coefficient
    beta3:float=0.99, # Gradient squared moving average (β3) coefficient
    eps:float=1e-8, # Added for numerical stability
    wd:float=0.02, # True weight decay
    paper_init:bool=False, # Initialize prior gradient with current gradient per paper, or zeroes
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|AdanForEachOptimizer|JitOptimizer:
    "Partial function for the Adan optimizer with fused ForEach and TorchScript implementations"
    return partialler(Adan, beta1=beta1, beta2=beta2, beta3=beta3, eps=eps, wd=wd,
                      paper_init=paper_init, foreach=foreach, jit=jit)

# %% ../../nbs/optimizer.adan.ipynb 20
def AdanLargeBatchLR(bs:int) -> float:
    "Square root rule for scaling `Adan` learning rate for large-batch training"
    return math.sqrt(bs/256)*6.25e-3



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/audio.05_mixup.ipynb.

# %% ../../nbs/audio.05_mixup.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/audio.05_mixup.ipynb 3
from __future__ import annotations

from torch.distributions import Bernoulli
from torch.distributions.beta import Beta

from fastcore.transform import Pipeline

from fastai.callback.mixup import reduce_loss
from fastai.layers import NoneReduce

from .data import MelSpectrogram, Spectrogram
from .augment import AmplitudeToDB, AudioNormalize
from ..multiloss import MixHandlerX
from ..imports import *

# %% auto 0
__all__ = ['AudioMixHandler', 'AudioMixUp', 'AudioCutMix', 'AudioCutMixUp', 'AudioCutMixUpAugment']

# %% ../../nbs/audio.05_mixup.ipynb 6
class AudioMixHandler(MixHandlerX):
    "Mixup base for `TensorAudio`"
    def __init__(self,
        alpha:float=0.5, # Alpha & beta parametrization for `Beta` distribution
        interp_label:bool|None=None # Blend or stack labels. Defaults to `loss_func.y_int` if None
    ):
        super().__init__(alpha, interp_label)

    def before_fit(self):
        waveforms, wave, spec = True, [], []
        self._wave_pipe = Pipeline([])
        self._spec_pipe = Pipeline([])

        # first copy transforms
        self._orig_pipe = self.dls.train.after_batch

        # loop through existing transforms appending to pre_spec/post_spec until Spec/Mel is found
        for i in range(len(self.dls.train.after_batch.fs)):
            if isinstance(self.dls.train.after_batch[i], (Spectrogram, MelSpectrogram)):
                waveforms = False
            if waveforms:
                wave.append(self.dls.train.after_batch[i])
            else:
                spec.append(self.dls.train.after_batch[i])

        self._wave_pipe.add(wave)
        self._spec_pipe.add(spec)

        # set existing transforms to an empty Pipeline
        self.dls.train.after_batch = Pipeline([])

    def after_fit(self):
        self.dls.train.after_batch = self._orig_pipe

    def after_cancel_fit(self):
        self.after_fit()
        super().after_cancel_fit()

# %% ../../nbs/audio.05_mixup.ipynb 8
class AudioMixUp(AudioMixHandler):
    "Implementation of https://arxiv.org/abs/1710.09412 for `TensorAudio`"
    def __init__(self, alpha=0.5, stack_y=True):
        super().__init__(alpha, stack_y)

    def before_batch(self,
        wave:bool=True # Apply waveform augmentations, used by `AudioCutMixAug`
    ):
        if wave: self.learn.xb = self._wave_pipe(self.xb)

        lam = self.distrib.sample((self.y.size(0),)).squeeze().to(self.x.device)
        lam = torch.stack([lam, 1-lam], 1)
        self.lam = lam.max(1)[0]
        shuffle = torch.randperm(self.y.size(0)).to(self.x.device)
        xb1,self.yb1 = tuple(L(self.xb).itemgot(shuffle)),tuple(L(self.yb).itemgot(shuffle))
        nx_dims = len(self.x.size())
        self.learn.xb = tuple(L(xb1,self.xb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=nx_dims-1)))

        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

        self.learn.xb = self._spec_pipe(self.xb)

# %% ../../nbs/audio.05_mixup.ipynb 10
class AudioCutMix(AudioMixHandler):
    "Implementation of https://arxiv.org/abs/1710.09412 for `TensorAudio`"
    def __init__(self,
        alpha:float=1., # Alpha & beta parametrization for `Beta` distribution
        uniform:bool=True, # Uniform patches across batch. True matches fastai CutMix
        p:float=1., # Per Image probablily of applying CutMix if `uniform` is False
        interp_label:bool|None=None # Blend or stack labels. Defaults to loss' `y_int` if None
    ):
        super().__init__(alpha, interp_label)
        store_attr(but='alpha,interp_label')
        if not uniform: self.bernoulli = Bernoulli(p)

    def before_batch(self,
        wave:bool=True # Apply waveform augmentations, used by `AudioCutMixAug`
    ):
        "Add clips and blend labels from another random item in batch"
        if wave:
            self.learn.xb = self._wave_pipe(self.xb)
        if self.uniform:
            self._uniform_cutmix()
        else:
            self._multi_cutmix()
        self.learn.xb = self._spec_pipe(self.xb)

    def _uniform_cutmix(self):
        "Add uniform clip and blend labels from another random item in batch"
        bs, _, X = self.x.size()
        self.lam = self.distrib.sample((1,)).to(self.x.device)
        shuffle = torch.randperm(bs).to(self.x.device)
        xb1,self.yb1 = self.x[shuffle], tuple((self.y[shuffle],))
        x1, x2 = self.rand_cut(X, self.lam)
        self.learn.xb[0][..., x1:x2] = xb1[..., x1:x2]
        self.lam = (1 - (x2-x1)/float(X))
        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

    def _multi_cutmix(self):
        "Add random clip and blend labels from another random item in batch"
        bs, _, X = self.x.size()
        samples = self.bernoulli.sample((bs,)).sum().int()
        idxes = torch.multinomial(torch.ones(bs, device=self.x.device), samples).to(self.x.device)
        self.lam = self.distrib.sample((samples,)).to(self.x.device)
        shuffle = torch.randperm(bs).to(self.x.device)
        xb1,self.yb1 = self.x[idxes][shuffle], tuple((self.y[idxes][shuffle],))
        for i, idx in enumerate(idxes):
            if 1 > self.lam[i] > 0:
                x1, x2 = self.rand_cut(X, self.lam[i])
                self.learn.xb[0][idx, ..., x1:x2] = xb1[i, ..., x1:x2]
                self.lam[i] = (1 - (x2-x1)/float(X))
        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

    def rand_cut(self,
        X:int, # Input audio length
        lam:Tensor # Lambda sample from Beta distribution
    ) -> tuple[Tensor, Tensor]:
        "Return random audio clip coordinates"
        cut_rat = torch.sqrt(1. - lam).to(self.x.device)
        cut_x = torch.round(X * cut_rat).type(torch.long).to(self.x.device)
        cut_x = torch.div(cut_x, 2, rounding_mode='floor')
        # uniform
        cx = torch.randint(0, X, (1,)).to(self.x.device)
        x1 = torch.clamp(cx - cut_x, 0, X)
        x2 = torch.clamp(cx + cut_x, 0, X)
        return x1, x2

# %% ../../nbs/audio.05_mixup.ipynb 12
class AudioCutMixUp(AudioMixUp, AudioCutMix):
    "Implementation of Mixup or CutMix for `TensorAudio`"
    def __init__(self, mix_alpha=.4, cut_alpha=1., stack_y=True, cut_ratio=1, mix_ratio=1): 
        AudioMixUp.__init__(self, mix_alpha, stack_y)
        AudioCutMix.__init__(self, cut_alpha, stack_y)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.ratio = mix_ratio / (cut_ratio + mix_ratio)

    def before_batch(self):
        if torch.rand(1) <= self.ratio: #mixup
            self.distrib = self.mix_distrib
            AudioMixUp.before_batch(self)
        else:
            self.distrib = self.cut_distrib
            AudioCutMix.before_batch(self)

# %% ../../nbs/audio.05_mixup.ipynb 14
class AudioCutMixUpAugment(AudioMixUp, AudioCutMix):
    "Implementation of Mixup, CutMix, or Augment for `TensorAudio`"
    def __init__(self, mix_alpha=.4, cut_alpha=1., stack_y=True, aug_ratio=1, cut_ratio=1, mix_ratio=1, augs_only=None, wave_augs=False): 
        AudioMixUp.__init__(self, mix_alpha, stack_y)
        AudioCutMix.__init__(self, cut_alpha, stack_y)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.aug_cutmix_ratio = aug_ratio / (aug_ratio + cut_ratio + mix_ratio)
        if self.aug_cutmix_ratio == 1: self.cut_mix_ratio = 0
        else: self.cut_mix_ratio = mix_ratio / (cut_ratio + mix_ratio)
        self.augs_only = augs_only
        self.wave_augs = wave_augs

    def before_fit(self):
        if self.augs_only is None: self.augs_only = (self.learn.n_epoch + 1)/self.learn.n_epoch
        elif self.augs_only >=1: self.augs_only = self.augs_only/self.learn.n_epoch
        else: self.augs_only = self.augs_only

        waveforms, wave, spec, norm = True, [], [], []
        self._wave_pipe = Pipeline([])
        self._spec_pipe = Pipeline([])
        self._norm_pipe = Pipeline([])
        
        # first copy transforms
        self._orig_pipe = self.dls.train.after_batch
        self._orig_pipe.split_idx = 0 # need to manually set split_idx for training augmentations to run
        
        # loop through existing transforms appending to pre_spec/post_spec until Spec/Mel is found
        for i in range(len(self.dls.train.after_batch.fs)):
            if isinstance(self.dls.train.after_batch[i], (Spectrogram, MelSpectrogram)):
                waveforms = False
            
            if waveforms: 
                wave.append(self.dls.train.after_batch[i])
            else:
                if isinstance(self.dls.train.after_batch[i], (AmplitudeToDB, AudioNormalize)):
                    norm.append(self.dls.train.after_batch[i])
                elif isinstance(self.dls.train.after_batch[i], (Spectrogram, MelSpectrogram)):
                    spec.append(self.dls.train.after_batch[i])

        self._wave_pipe.add(wave)
        self._spec_pipe.add(spec)
        self._norm_pipe.add(norm)

        # set existing transforms to an empty Pipeline
        self.dls.train.after_batch = Pipeline([])

    def before_batch(self):
        if self.augs_only >= self.learn.pct_train and torch.rand(1) >= self.aug_cutmix_ratio: # augs or mixup/cutmix
            self._doaugs = False
            if self.cut_mix_ratio > 0 and torch.rand(1) <= self.cut_mix_ratio: # mixup or cutmix
                self.distrib = self.mix_distrib
                AudioMixUp.before_batch(self, self.wave_augs)
            else:
                self.distrib = self.cut_distrib
                AudioCutMix.before_batch(self, self.wave_augs)
            self.learn.xb = self._norm_pipe(self.xb) # now normalize
        else:
            self._doaugs = True
            self.learn.xb = self._orig_pipe(self.xb) # original transforms

    def after_cancel_fit(self):
        self.after_fit()
        AudioMixUp.after_cancel_fit(self)

    def solo_lf(self, pred, *yb):
        "`norm_lf` applies the original loss function on both outputs based on `self.lam` if applicable"
        if not self.training or self._doaugs: 
            return self.old_lf(pred, *yb)
        with NoneReduce(self.old_lf) as lf:
            loss = torch.lerp(lf(pred,*self.yb1), lf(pred,*yb), self.lam)
        return reduce_loss(loss, getattr(self.old_lf, 'reduction', 'mean'))

    def multi_lf(self, pred, *yb):
        "`norm_lf` applies the original loss function on both outputs based on `self.lam` if applicable"
        if not self.training or self._doaugs:
            return self.learn.loss_func_mixup(pred, *yb)
        else:
            return self.learn.loss_func_mixup.forward_mixup(pred, *self.yb1, *yb, self.lam)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/vision.augment.itemtensor.ipynb.

# %% ../../../nbs/vision.augment.itemtensor.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../../nbs/vision.augment.itemtensor.ipynb 3
from __future__ import annotations

import operator
from PIL import Image
from torchvision.transforms.functional import pad as tvpad
from torchvision.transforms.functional import _interpolation_modes_from_int
from torch.nn.functional import interpolate

from fastai.vision.augment import RandomCrop, CropPad, Resize, PadMode, RandomResizedCrop, RatioResize, ResizeMethod, _pad_modes, _get_sz
from fastai.vision.core import TensorImage, TensorMask

from ...imports import *

# %% auto 0
__all__ = ['encodes']

# %% ../../../nbs/vision.augment.itemtensor.ipynb 7
def _resize(
    x:TensorImage|TensorMask,
    size:int|tuple[int,...],
    shape:list[int]|tuple[int,...],
    interpolation:str
):
    align_corners = False if interpolation in ["bilinear", "bicubic"] else None
    return interpolate(x.view(shape), size=size, mode=interpolation, align_corners=align_corners)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 8
@patch
def resize(x:TensorImage|TensorMask,
    size:int|tuple[int,int],
    interpolation:Enum
):
    if len(x.shape)==3:
        l, c, h, w = 3, x.shape[0], x.shape[1], x.shape[2]
    elif len(x.shape)==2:
        l, c, h, w = 2, 1, x.shape[0], x.shape[1]

    if len(size)==2:
        sh, sw = size[0], size[1]
    elif len(size)==1:
        sh, sw = size, size

    x = _resize(x, size=size, shape=[1,c,h,w], interpolation=interpolation.value)

    if l==3:
        return x.view(c, sh, sw)
    else:
        return x.view(sh, sw)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 9
@patch
def _do_crop_pad(x:TensorImage|TensorMask,
    sz:fastuple, # Crop/pad size of input
    tl:fastuple, # Top-left coordinate of the crop/pad, if `None` center crop
    orig_sz:fastuple, # Original size of input
    pad_mode:PadMode=PadMode.Zeros, # Fastai padding mode
    resize_mode:int=Image.BILINEAR, # Pillow `Image` resize mode
    resize_to:tuple[int,int]|None=None # Post crop/pad resize of input
):
    # PyTorch and PIL axis are opposite, need to reverse PIL axis input for crop and resize
    if any(tl.ge(0)) or any(tl.add(sz).le(orig_sz)):
        # At least one dim is inside the image, so needs to be cropped
        c = tl.max(0)
        left, top, right, bottom = *c, *tl.add(sz).min(orig_sz)
        x = x[..., top:bottom, left:right]
    if any(tl.lt(0)) or any(tl.add(sz).ge(orig_sz)):
        # At least one dim is outside the image, so needs to be padded
        p = (-tl).max(0)
        f = (sz-orig_sz).add(tl).max(0)
        if len(x.shape)==2:
            x = x.view(1, x.shape[0], x.shape[1])
            x = tvpad(x, (*p, *f), padding_mode=_pad_modes[pad_mode])
            x = x.view(x.shape[1], x.shape[2])
        else:
            x = tvpad(x, (*p, *f), padding_mode=_pad_modes[pad_mode])
    if resize_to is not None:
        resize_mode = Image.NEAREST if isinstance(x,TensorMask) else resize_mode
        x = x.resize([*resize_to][::-1], _interpolation_modes_from_int(resize_mode))
    return x

# %% ../../../nbs/vision.augment.itemtensor.ipynb 10
@patch
def crop_pad(x:TensorImage|TensorMask,
    sz:int|tuple[int,int], # Crop/pad size of input, duplicated if one value is specified
    tl:tuple[int,int]|None=None, # Top-left coordinate of the crop/pad, if `None` center crop
    orig_sz:tuple[int,int]|None=None, # Original size of input
    pad_mode:PadMode=PadMode.Zeros, # Fastai padding mode
    resize_mode:int=Image.BILINEAR, # Pillow `Image` resize mode
    resize_to:tuple[int,int]|None=None # Post crop/pad resize of input
):
    if isinstance(sz,int):
        sz = (sz,sz)
    orig_sz = fastuple(_get_sz(x) if orig_sz is None else orig_sz)
    sz,tl = fastuple(sz),fastuple(((_get_sz(x)-sz)//2) if tl is None else tl)
    if isinstance(x,TensorMask):
        return x.float()._do_crop_pad(sz, tl, orig_sz=orig_sz, pad_mode=pad_mode, resize_mode=resize_mode, resize_to=resize_to).long()
    else:
        return x._do_crop_pad(sz, tl, orig_sz=orig_sz, pad_mode=pad_mode, resize_mode=resize_mode, resize_to=resize_to)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 12
@RandomCrop
def encodes(self, x:TensorImage|TensorMask):
    return x.crop_pad(self.size, self.tl, orig_sz=self.orig_sz)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 18
@CropPad
def encodes(self, x:TensorImage|TensorMask):
    orig_sz = _get_sz(x)
    tl = (orig_sz-self.size)//2
    return x.crop_pad(self.size, tl, orig_sz=orig_sz, pad_mode=self.pad_mode)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 22
@Resize
def encodes(self, x:TensorImage|TensorMask):
    orig_sz = _get_sz(x)
    if self.method==ResizeMethod.Squish:
        return x.crop_pad(orig_sz, fastuple(0,0), orig_sz=orig_sz, pad_mode=self.pad_mode,
                resize_mode=self.mode_mask if isinstance(x,TensorMask) else self.mode, resize_to=self.size)

    w,h = orig_sz
    op = (operator.lt,operator.gt)[self.method==ResizeMethod.Pad]
    m = w/self.size[0] if op(w/self.size[0],h/self.size[1]) else h/self.size[1]
    cp_sz = (int(m*self.size[0]),int(m*self.size[1]))
    tl = fastuple(int(self.pcts[0]*(w-cp_sz[0])), int(self.pcts[1]*(h-cp_sz[1])))
    return x.crop_pad(cp_sz, tl, orig_sz=orig_sz, pad_mode=self.pad_mode,
                resize_mode=self.mode_mask if isinstance(x,TensorMask) else self.mode, resize_to=self.size)

# %% ../../../nbs/vision.augment.itemtensor.ipynb 28
@RandomResizedCrop
def encodes(self, x:TensorImage|TensorMask):
    res = x.crop_pad(self.cp_size, self.tl, orig_sz=self.orig_sz,
        resize_mode=self.mode_mask if isinstance(x,TensorMask) else self.mode, resize_to=self.final_size)
    if self.final_size != self.size: res = res.crop_pad(self.size) #Validation set: one final center crop
    return res

# %% ../../../nbs/vision.augment.itemtensor.ipynb 34
@RatioResize
def encodes(self, x:TensorImage|TensorMask):
    w,h = _get_sz(x)
    if w >= h: nw,nh = self.max_sz,h*self.max_sz/w
    else:      nw,nh = w*self.max_sz/h,self.max_sz
    return Resize(size=(int(nh),int(nw)), resamples=self.resamples)(x)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/optimizer.finetune.ipynb.

# %% ../../nbs/optimizer.finetune.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/optimizer.finetune.ipynb 4
from __future__ import annotations

from fastcore.basics import GetAttr

from fastai.optimizer import (Optimizer, _update, weight_decay, l2_reg, average_grad, sgd_step,
                              momentum_step, average_sqr_grad, rms_prop_step, step_stat, adam_step,
                              radam_step, qhadam_step, larc_layer_lr, larc_step, lamb_step)
from ..imports import *

# %% auto 0
__all__ = ['FineTuneOpt', 'fine_tune_wd', 'SGDFT', 'RMSPropFT', 'AdamFT', 'RAdamFT', 'QHAdamFT', 'LarcFT', 'LambFT',
           'LookaheadFT', 'rangerFT']

# %% ../../nbs/optimizer.finetune.ipynb 8
class FineTuneOpt(Optimizer):
    """
    Modification of the base optimizer class for the fastai library, updating `params` with `cbs`

    In combination with the `fine_tune_wd` callback, adds optional weight decay `ft_wd` towards the starting value, 
    to prevent overfitting to the new dataset during fine-tuning.

    By default, will not apply to the fine-tuning head, just the pretrained body.

    From: https://gist.github.com/crowsonkb/f646976de8033b371ea17cb9b1c1561f
    """
    _keep_on_clear = ['force_train', 'do_wd']
    def __init__(self,
        params:Tensor, # Parameters and hyper parameters
        cbs:list, # `Optimizer` callbacks
        train_bn:bool=True, # Batch normalization is always trained
        wd_ft_head:bool=False, # Apply fine tuning weight decay to model head
        **defaults # Default values to set on hyper parameters
    ):
        super().__init__(params, cbs, train_bn, **defaults)
        self.wd_ft_head, self.set_orig_p = wd_ft_head, True

    @torch.no_grad()
    def step(self, closure=None):
        if self.set_orig_p:
            self.set_orig_p = False
            n = slice(None) if self.wd_ft_head or len(self.param_lists)<=1 else slice(None, -1)
            for p,pg,state,hyper in self.all_params(n):
                state['orig_p'] = p.detach().clone()
                self.state[p] = state
        super().step(closure)

    def clear_state(self):
        self.set_orig_p = True
        super().clear_state()

    def state_dict(self):
        state = super().state_dict()
        state.update({'set_orig_p': self.set_orig_p})
        return state

    def load_state_dict(self, sd):
        self.set_orig_p = sd.pop('set_orig_p')
        super().load_state_dict(sd)

# %% ../../nbs/optimizer.finetune.ipynb 12
def fine_tune_wd(p, lr, ft_wd, orig_p=None, do_wd=True, **kwargs):
    "Weight decay `p` towards the starting value `orig_p`"
    if do_wd and ft_wd !=0 and orig_p is not None:
        p.lerp_(orig_p, lr*ft_wd)

# %% ../../nbs/optimizer.finetune.ipynb 15
def SGDFT(params, lr, mom=0., wd=0., ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "A `Optimizer` for SGD with `lr` and `mom` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd]
    if mom != 0: cbs.append(average_grad)
    cbs.append(sgd_step if mom==0 else momentum_step)
    return FineTuneOpt(params, cbs, lr=lr, mom=mom, wd=wd, ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 19
def RMSPropFT(params, lr, sqr_mom=0.99, mom=0., wd=0., ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "A `FineTuneOpt` for RMSProp with `lr`, `sqr_mom`, `mom` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd] + [average_sqr_grad] if mom==0. else [average_grad, average_sqr_grad]
    cbs.append(rms_prop_step)
    return FineTuneOpt(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, wd=wd, ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 21
def AdamFT(params, lr, mom=0.9, sqr_mom=0.99, eps=1e-5, wd=0.01, ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "A `FineTuneOpt` for Adam with `lr`, `mom`, `sqr_mom`, `eps` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd, partial(average_grad, dampening=True), average_sqr_grad, step_stat, adam_step]
    return FineTuneOpt(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 23
def RAdamFT(params, lr, mom=0.9, sqr_mom=0.99, eps=1e-5, wd=0., ft_wd=0., beta=0., decouple_wd=True, wd_ft_head=False):
    "A `FineTuneOpt` for Adam with `lr`, `mom`, `sqr_mom`, `eps` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd, partial(average_grad, dampening=True), average_sqr_grad, step_stat, radam_step]
    return FineTuneOpt(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, ft_wd=ft_wd, beta=beta, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 25
def QHAdamFT(params, lr, mom=0.999, sqr_mom=0.999, nu_1=0.7, nu_2 = 1.0, eps=1e-8, wd=0., ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "An `FineTuneOpt` for Adam with `lr`, `mom`, `sqr_mom`, `nus`, eps` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd, partial(average_grad, dampening=True), partial(average_sqr_grad, dampening=True), step_stat, qhadam_step]
    return FineTuneOpt(params, cbs, lr=lr, nu_1=nu_1, nu_2=nu_2 ,
                       mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd,
                       ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 27
def LarcFT(params, lr, mom=0.9, clip=True, trust_coeff=0.02, eps=1e-8, wd=0., ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "A `FineTuneOpt` for Adam with `lr`, `mom`, `sqr_mom`, `eps` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd]
    if mom!=0.: cbs.append(average_grad)
    cbs += [partial(larc_layer_lr, clip=clip), larc_step]
    return FineTuneOpt(params, cbs, lr=lr, mom=mom, trust_coeff=trust_coeff, eps=eps, wd=wd, ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 29
def LambFT(params, lr, mom=0.9, sqr_mom=0.99, eps=1e-5, wd=0., ft_wd=0., decouple_wd=True, wd_ft_head=False):
    "A `FineTuneOpt` for Adam with `lr`, `mom`, `sqr_mom`, `eps` and `params`"
    cbs = [weight_decay] if decouple_wd else [l2_reg]
    cbs += [fine_tune_wd, partial(average_grad, dampening=True), average_sqr_grad, step_stat, lamb_step]
    return Optimizer(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, ft_wd=ft_wd, wd_ft_head=wd_ft_head)

# %% ../../nbs/optimizer.finetune.ipynb 31
class LookaheadFT(FineTuneOpt, GetAttr):
    "Wrap a `FineTuneOpt` `opt` in a Lookahead optimizer"
    _default='opt'
    def __init__(self, opt, k=6, alpha=0.5):
        store_attr('opt,k,alpha')
        self._init_state()

    def step(self, closure=None):
        if closure is not None: raise NotImplementedError("fastai optimizers currently do not support closure")
        if self.slow_weights is None: self._copy_weights()
        self.opt.step()
        self.count += 1
        if self.count%self.k != 0: return
        for slow_pg,fast_pg in zip(self.slow_weights,self.param_lists):
            for slow_p,fast_p in zip(slow_pg,fast_pg):
                slow_p.data.add_(fast_p.data-slow_p.data, alpha=self.alpha)
                fast_p.data.copy_(slow_p.data)

    def clear_state(self):
        self.opt.clear_state()
        self._init_state()

    def state_dict(self):
        state = self.opt.state_dict()
        state.update({'count': self.count, 'slow_weights': self.slow_weights})
        return state

    def load_state_dict(self, sd):
        self.count = sd.pop('count')
        self.slow_weights = sd.pop('slow_weights')
        self.opt.load_state_dict(sd)

    def _init_state(self): self.count,self.slow_weights = 0,None
    def _copy_weights(self): self.slow_weights = L(L(p.clone().detach() for p in pg) for pg in self.param_lists)

    @property
    def param_lists(self): return self.opt.param_lists
    @param_lists.setter
    def param_lists(self, v): self.opt.param_lists = v

# %% ../../nbs/optimizer.finetune.ipynb 33
@delegates(RAdamFT)
def rangerFT(p, lr, mom=0.95, wd=0.01, ft_wd=0., eps=1e-6, **kwargs):
    "Convenience method for `LookaheadFT` with `RAdamFT`"
    return LookaheadFT(RAdamFT(p, lr=lr, mom=mom, wd=wd, ft_wd=ft_wd, eps=eps, **kwargs))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/audio.03_augment.ipynb.

# %% ../../nbs/audio.03_augment.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/audio.03_augment.ipynb 3
from __future__ import annotations

import random
import math
from collections import Counter
from fractions import Fraction
from functools import reduce
from itertools import chain, count, islice, repeat

import colorednoise
from primePy import primes

from torch import _VF
from torch.distributions import Bernoulli
import torchaudio.transforms as tatfms
import torchaudio.functional as TAF
from torchaudio.functional.functional import _get_sinc_resample_kernel, _apply_sinc_resample_kernel

from fastcore.transform import DisplayedTransform, retain_type

from fastai.data.load import DataLoader
from fastai.vision.augment import RandTransform, broadcast_vec

from ..transform import BatchRandTransform
from .core import TensorAudio, TensorSpec, TensorMelSpec
from ..imports import *

# %% auto 0
__all__ = ['Flip', 'Roll', 'AudioPadMode', 'RandomCropPad', 'VolumeMode', 'Volume', 'PeakNorm', 'VolumeOrPeakNorm', 'Noise',
           'VolumeBatch', 'PitchShift', 'PitchShiftTA', 'TimeStretch', 'PitchShiftOrTimeStretch', 'TimeMasking',
           'FrequencyMasking', 'AmplitudeToDBMode', 'AmplitudeToDB', 'AudioNormalize']

# %% ../../nbs/audio.03_augment.ipynb 6
def _get_audio_attr(x:TensorAudio|TensorSpec|TensorMelSpec|tuple[TensorAudio|TensorSpec|TensorMelSpec,...], attr:str) -> Any:
    if isinstance(x, tuple):
        return getattr(x[0], attr)
    else:
        return getattr(x, attr)

# %% ../../nbs/audio.03_augment.ipynb 8
class Flip(RandTransform):
    "Randomly flip `TensorAudio` with probability `p`"
    order, split_idx = 5, 0
    def __init__(self, p:float=0.5):
        super().__init__(p=p)

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.fliplr()

# %% ../../nbs/audio.03_augment.ipynb 11
class Roll(RandTransform):
    "Randomly shift `TensorAudio` with rollover"
    order, split_idx = 6, 0
    def __init__(self,
        p:float=0.5, # Per-item probability
        max_roll:float=0.5, # Maximum shift
    ):
        store_attr('max_roll')
        super().__init__(p=p)

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0: self.shift = int(random.uniform(-1,1) * self.max_roll * _get_audio_attr(b, 'samples'))

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.roll(shifts=self.shift, dims=-1)

# %% ../../nbs/audio.03_augment.ipynb 14
class AudioPadMode(Enum):
    "All methods of padding audio as attributes to get tab-completion and typo-proofing",
    Constant = 1
    ConstantPre = 2
    ConstantPost = 3
    Repeat = 4

# %% ../../nbs/audio.03_augment.ipynb 15
@patch
def crop_pad(x:TensorAudio,
    end_len:int,
    crop_start:int|None=None,
    pad_len:int=0,
    padmode:AudioPadMode=AudioPadMode.Repeat,
    constant:float=0
):
    if crop_start is not None:
        x = x[...,crop_start:crop_start+end_len]
    elif pad_len:
        if padmode==AudioPadMode.ConstantPre:
            x = _VF.constant_pad_nd(x, (pad_len, 0), constant)
        elif padmode==AudioPadMode.ConstantPost:
            x = _VF.constant_pad_nd(x, (0, pad_len), constant)
        elif padmode==AudioPadMode.Repeat:
            if pad_len >= x.samples:
                x = x.repeat(1, 2+max(pad_len//x.samples, 2))
            else:
                x = torch.cat([x, x[...,0:pad_len]], dim=-1)
        elif padmode==AudioPadMode.Constant:
            if pad_len % 2 == 0: pad = pad_len//2
            else:                pad = pad_len//2 + 1
            x = _VF.constant_pad_nd(x, (pad, pad), constant)

    if x.samples > end_len:
        x = x[...,0:end_len]
    return x

# %% ../../nbs/audio.03_augment.ipynb 16
class RandomCropPad(RandTransform):
    "Randomly resize `TensorAudio` to specified length. Center crop during valid."
    split_idx, order = None, 25
    def __init__(self,
        duration:Numeric|None=None, # Crop length in seconds
        samples:int|None=None, # Crop length in samples
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Numeric=0, # Value for `AudioPadMode.Constant`
    ):
        if duration is None and samples is None:
            raise ValueError('One of `duration` or `samples` must be set')
        if constant >= 1 or constant <= -1:
            raise ValueError(f'{constant=}` must be between [-1,1]')
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"{padmode=} is not valid")
        store_attr()
        super().__init__()

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        self.samples = self.samples if self.samples is not None else int(self.duration*_get_audio_attr(b, 'sr'))
        self.orig_samples = _get_audio_attr(b, 'samples')
        if split_idx:
            self.crop_start = (self.orig_samples-self.samples)//2 if self.samples < self.orig_samples else None
        else:
            self.crop_start = torch.randint(0,self.orig_samples-self.samples, (1,)) if self.samples < self.orig_samples else None

        self.pad_len = (self.samples-self.orig_samples) if self.samples > self.orig_samples else 0

    def encodes(self, x:TensorAudio) -> Tensor:
        return x.crop_pad(self.samples, self.crop_start, self.pad_len, self.padmode, self.constant)

# %% ../../nbs/audio.03_augment.ipynb 24
class VolumeMode(Enum):
    "All Volume modes as attributes to get tab-completion and typo-proofing",
    DB = 1
    Power = 2
    Amplitude = 3

# %% ../../nbs/audio.03_augment.ipynb 25
class Volume(RandTransform):
    "Randomly change `TensorAudio`'s volume"
    order, split_idx = 30, 0
    def __init__(self,
        p:float=0.75, # Per-item probability
        gain:Numeric|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Numeric,Numeric]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"{volmode=} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0 and self._random_gain: self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.volmode == VolumeMode.DB:
            x = TAF.gain(x, self.gain)
        elif self.volmode == VolumeMode.Amplitude:
            x = x * self.gain
        elif self.volmode == VolumeMode.Power:
            x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# %% ../../nbs/audio.03_augment.ipynb 28
class PeakNorm(RandTransform):
    "Randomly apply constant gain so `TensorAudio`'s loudest level is between -1 and 1 with probability `p`"
    order, split_idx = 31, 0
    def __init__(self, p:float=0.1):
        super().__init__(p=p)
        store_attr(but='p')

    def encodes(self, x:TensorAudio) -> Tensor:
            return x / torch.max(torch.abs(x))

# %% ../../nbs/audio.03_augment.ipynb 31
class VolumeOrPeakNorm(RandTransform):
    "Randomly apply `Volume` or `Peak` to `TensorAudio` in one transform"
    order, split_idx = 30, 0
    def __init__(self,
        p:float=0.75, # Per-item probability
        peak_p:float=0.1, # Probability of applying `Peak`
        gain:Numeric|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Numeric,Numeric]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"{volmode=} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0:
            self.peak = torch.randn(1) < self.peak_p
            if not self.peak and self._random_gain:
                self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.peak:
            x = x / torch.max(torch.abs(x))
        else:
            if self.volmode == VolumeMode.DB:
                x = TAF.gain(x, self.gain)
            elif self.volmode == VolumeMode.Amplitude:
                x = x * self.gain
            elif self.volmode == VolumeMode.Power:
                x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# %% ../../nbs/audio.03_augment.ipynb 33
# Noise contains code from https://github.com/fastaudio/fastaudio:
# fastaudio - MIT License - Copyright (c) 2020 scart97

class NoiseColor(Enum):
    """Exponent beta values for named noise colors (see `colored_noise`)."""
    Violet = -2
    Blue = -1
    White = 0
    Pink = 1
    Brown = 2

# %% ../../nbs/audio.03_augment.ipynb 34
class Noise(RandTransform):
    "Adds noise of specified color and level to `TensorAudio` relative to mean audio level"
    order, split_idx = 60, 0
    def __init__(self,
        p=0.25, # Per-item probability
        noise_level:float|None=None, # Loudness of noise, if None randomly selects from `noise_range`
        noise_range:tuple[float,float]=(0.01,0.1), # Range of noise loudness values
        color:NoiseColor|None=None, # Color of noise to add, if None randomly selects from `NoiseColor`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self._color_choice = [nc for nc in NoiseColor]
        if color is not None and color not in self._color_choice:
            raise ValueError(f"{color=} is not valid")
        self._random_color = color is None
        self._random_noise = noise_level is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0:
            if self._random_color: self.color = random.choice(self._color_choice)
            if self._random_noise: self.noise_level = random.uniform(*self.noise_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        # if it's white noise, use torch.randn_like for speed
        if self.color == NoiseColor.White:
            noise = torch.randn_like(x)
        else:
            noise = torch.from_numpy(colorednoise.powerlaw_psd_gaussian(exponent=self.color.value, size=x.samples)).float()
        x += noise * x.abs().mean() * self.noise_level
        return x

# %% ../../nbs/audio.03_augment.ipynb 38
class VolumeBatch(BatchRandTransform):
    "Randomly change `TensorAudio`'s volume on the GPU"
    order, split_idx = 10, 0
    def __init__(self,
        p:float=0.5, # Per-item probability
        gain:Numeric|None=None, # Gain is a positive amplitude ratio, a power (voltage squared), or in decibels. If none, randomly select from `gain_range`.
        gain_range:tuple[Numeric,Numeric]=(-18, 6), # Random range for gain
        volmode:VolumeMode=VolumeMode.DB # One of `VolumeMode.DB`, `VolumeMode.Amplitude`, or `VolumeMode.Power`
    ):
        super().__init__(p=p)
        store_attr(but='p')
        if volmode not in [vol for vol in VolumeMode]:
            raise ValueError(f"{volmode=} is not valid")
        if volmode != VolumeMode.DB:
            self.gain_range = (max(gain_range[0],0), gain_range[1])
        self._random_gain = gain is None

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if split_idx==0 and self._random_gain: self.gain = random.uniform(*self.gain_range)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.volmode == VolumeMode.DB:
            x = TAF.gain(x, self.gain)
        elif self.volmode == VolumeMode.Amplitude:
            x = x * self.gain
        elif self.volmode == VolumeMode.Power:
            x = TAF.gain(x, 10 * math.log10(self.gain))

        return torch.clamp(x, -1, 1)

# %% ../../nbs/audio.03_augment.ipynb 40
# PitchShift & TimeStretch contains code from https://github.com/KentoNishi/torch-pitch-shift & https://github.com/KentoNishi/torch-time-stretch
# torch-pitch-shift & torch-time-stretch - MIT License - Copyright (c) 2021 Kento Nishi

# %% ../../nbs/audio.03_augment.ipynb 41
# https://stackoverflow.com/a/46623112/9325832
def _combinations_without_repetition(r, iterable=None, values=None, counts=None):
    if iterable:
        values, counts = zip(*Counter(iterable).items())

    f = lambda i, c: chain.from_iterable(map(repeat, i, c))
    n = len(counts)
    indices = list(islice(f(count(), counts), r))
    if len(indices) < r:
        return
    while True:
        yield tuple(values[i] for i in indices)
        for i, j in zip(reversed(range(r)), f(reversed(range(n)), reversed(counts))):
            if indices[i] != j:
                break
        else:
            return
        j = indices[i] + 1
        for i, j in zip(range(i, r), f(count(j), counts[j:])):
            indices[i] = j

# %% ../../nbs/audio.03_augment.ipynb 42
def _get_fast_shifts(
    sample_rate:int, # Audio sample rate
    condition:Callable[...,bool]=lambda x: x >= 0.5 and x <= 2 and x != 1, # Function to validate fast shift ratios
) -> list[Fraction]: # Fast pitch-shift target ratios
    "Search for pitch-shift targets that can be computed quickly for a given sample rate"
    fast_shifts = set()
    factors = primes.factors(sample_rate)
    products = []
    for i in range(1, len(factors) + 1):
        products.extend(
            [
                reduce(lambda x, y: x * y, x)
                for x in _combinations_without_repetition(i, iterable=factors)
            ]
        )
    for i in products:
        for j in products:
            f = Fraction(i, j)
            if condition(f):
                fast_shifts.add(f)
    return list(fast_shifts)

# %% ../../nbs/audio.03_augment.ipynb 43
def _semitones_to_ratio(semitones:float) -> Fraction:
    "Convert semitonal shifts into a Fraction indicating a pitch shift ratio"
    return Fraction(2.0 ** (semitones / 12.0))

# %% ../../nbs/audio.03_augment.ipynb 44
def _get_fast_stretches(
    sample_rate:int, # Audio sample rate
    condition:Callable[...,bool]=lambda x: x >= 0.5 and x <= 2 and x != 1, # Function to validate fast stretch ratios.
) -> list[Fraction]: # Fast time-stretch target ratios
    "Search for time-stretch targets that can be computed quickly for a given sample rate"
    fast_shifts = set()
    factors = primes.factors(sample_rate)
    products = []
    for i in range(1, len(factors) + 1):
        products.extend(
            [
                reduce(lambda x, y: x * y, x)
                for x in _combinations_without_repetition(i, iterable=factors)
            ]
        )
    for i in products:
        for j in products:
            f = Fraction(i, j)
            if condition(f):
                fast_shifts.add(f)
    return list(fast_shifts)

# %% ../../nbs/audio.03_augment.ipynb 46
@patch
def pitch_shift(x:TensorAudio, n_fft, hop_length, shift, sr, new_sr, gcd, kernel, width, padmode, constant, device):
    shape = x.shape
    x = x.reshape(shape[0] * shape[1], shape[2])
    x = torch.stft(x, n_fft, hop_length, return_complex=True)
    phase_advance = torch.linspace(0, math.pi * hop_length, x.shape[1], device=device)[..., None]
    x = TAF.phase_vocoder(x, float(1 / shift), phase_advance)
    phase_advance = None
    x = torch.istft(x, n_fft, hop_length)
    x = retain_type(_apply_sinc_resample_kernel(x, sr, new_sr, gcd, kernel, width), typ=TensorAudio)
    crop_start = torch.randint(0, x.shape[-1]-shape[-1], (1,)) if shape[-1] < x.shape[-1]  else None
    pad_len = (shape[-1]-x.shape[-1]) if shape[-1] > x.shape[-1] else 0
    return x.crop_pad(shape[-1], crop_start, pad_len, padmode, constant).reshape(*shape)

# %% ../../nbs/audio.03_augment.ipynb 47
class PitchShift(BatchRandTransform):
    "Fast shift of `TensorAudio`'s pitch."
    order, split_idx = 20, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        semitones:tuple[float,float]=(-4.0,4.0), # Random pitch shift range in semitones to compute efficient shifts
        bins_per_octave:int=12, # Number of steps per octave
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Numeric=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent GPU OOM.
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self.sr = 0
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"{padmode=} is not valid")

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr:
            self.sr = _get_audio_attr(b, 'sr')
            self.n_fft = self.sr // 64
            self.hop_length = self.n_fft // 32
            self._fast_shifts = _get_fast_shifts(self.sr, self._validate_fast_shifts)
            self.type = _get_audio_attr(b, 'dtype')
            self.device = _get_audio_attr(b, 'device')
        if split_idx==0:
            self.shift = random.choice(self._fast_shifts)
            if not isinstance(self.shift, Fraction):
                self.shift = 2.0 ** (float(self.shift) / self.bins_per_octave)
            self.new_sr = int(self.sr/self.shift)
            self.gcd = math.gcd(self.sr, self.new_sr)
            self.kernel, self.width = _get_sinc_resample_kernel(self.sr, self.new_sr, self.gcd,
                                                                6, 0.99, 'sinc_interpolation',
                                                                None, self.device, self.type)

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.split:
            xs = []
            for xi in retain_type(x.split(self.split), TensorAudio):
                xs.append(xi.pitch_shift(self.n_fft, self.hop_length, self.shift, self.sr, self.new_sr,
                                         self.gcd, self.kernel, self.width, self.padmode, self.constant, self.device))
            return torch.cat(xs, dim=0)
        else:
            return x.pitch_shift(self.n_fft, self.hop_length, self.shift, self.sr, self.new_sr,
                                 self.gcd, self.kernel, self.width, self.padmode, self.constant, self.device)

    def _validate_fast_shifts(self, x):
        return x >= _semitones_to_ratio(self.semitones[0]) and x <= _semitones_to_ratio(self.semitones[1]) and x != 1

# %% ../../nbs/audio.03_augment.ipynb 50
class PitchShiftTA(BatchRandTransform):
    "Shift the `TensorAudio`'s pitch using TorchAudio. Can be slower than `PitchShift`"
    order, split_idx = 20, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        n_steps:int|None=None, # The (fractional) steps to shift waveform
        n_step_range:tuple[int,int] = (2,6), # Random `n_steps` range if `n_steps` is None
        bins_per_octave:int=12, # Number of steps per octave
        n_fft:int=512, # Size of FFT, creates n_fft // 2 + 1 bins
        win_length:int|None=None, # Window size. Defaults to `n_fft`
        hop_length:int|None=None, # Length of hop between STFT windows. Defaults to `win_length` // 4
        window_fn:Callable[..., Tensor]=torch.hann_window, # Window tensor applied/multiplied to each frame/window
        wkwargs:dict|None=None, # Args for `window_fn`
    ):
        super().__init__(p=p)
        store_attr(but='p,window_fn,wkwargs')
        self.sr=0
        self.win_length = win_length if win_length is not None else n_fft
        self.hop_length = hop_length if hop_length is not None else self.win_length // 4
        self.window = window_fn(self.win_length) if wkwargs is None else window_fn(self.win_length, **wkwargs)

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr:
            self.sr = _get_audio_attr(b, 'sr')
        if split_idx==0:
            self.steps = random.uniform(*self.n_step_range) if self.n_steps is None else self.n_steps

    def encodes(self, x:TensorAudio) -> Tensor:
        return TAF.pitch_shift(x, self.sr, self.steps, self.bins_per_octave, self.n_fft,
                               self.win_length, self.hop_length, self.window)

# %% ../../nbs/audio.03_augment.ipynb 53
@patch
def time_stretch(x:TensorAudio, n_fft, hop_length, stretch, padmode, constant, device):
    shape = x.shape
    x = x.reshape(shape[0] * shape[1], shape[2])
    x = torch.stft(x, n_fft, hop_length, return_complex=True)
    phase_advance = torch.linspace(0, math.pi * hop_length, x.shape[1], device=device)[..., None]
    x = TAF.phase_vocoder(x, float(1/stretch), phase_advance)
    phase_advance = None
    x = torch.istft(x, n_fft, hop_length)
    crop_start = torch.randint(0, x.shape[-1]-shape[-1], (1,)) if shape[-1] < x.shape[-1]  else None
    pad_len = (shape[-1]-x.shape[-1]) if shape[-1] > x.shape[-1] else 0
    return x.crop_pad(shape[-1], crop_start, pad_len, padmode, constant).reshape(*shape)

# %% ../../nbs/audio.03_augment.ipynb 54
class TimeStretch(BatchRandTransform):
    "Fast time stretch of `TensorAudio`"
    order, split_idx = 25, 0
    def __init__(self,
        p:float=0.2, # Per-item probability
        stretch_rates:tuple[float,float]=(0.5,2.0), # Random time stretch range to compute efficient stretches. Defaults to 50%-200% speed
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Numeric=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent GPU OOM.
    ):
        super().__init__(p=p)
        store_attr(but='p')
        self.sr = 0
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"{padmode=} is not valid")

    def before_call(self,
        b:TensorAudio|tuple[TensorAudio,...],
        split_idx:int # Index of the train/valid dataset
    ):
        super().before_call(b, split_idx)
        if not self.sr:
            self.sr = _get_audio_attr(b, 'sr')
            self.n_fft = self.sr // 64
            self.hop_length = self.n_fft // 32
            self._fast_stretchs = _get_fast_stretches(self.sr, self._validate_fast_stretches)
            self.device = _get_audio_attr(b, 'device')
        if split_idx==0:
            self.stretch = float(1/random.choice(self._fast_stretchs))

    def encodes(self, x:TensorAudio) -> Tensor:
        if self.split:
            xs = []
            for xi in retain_type(x.split(self.split), TensorAudio):
                xs.append(xi.time_stretch(self.n_fft, self.hop_length, self.stretch, self.padmode, self.constant, self.device))
            return torch.cat(xs, dim=0)
        else:
            return x.time_stretch(self.n_fft, self.hop_length, self.stretch, self.padmode, self.constant, self.device)

    def _validate_fast_stretches(self, x):
        return x >= self.stretch_rates[0] and x <= self.stretch_rates[1] and x != 1

# %% ../../nbs/audio.03_augment.ipynb 57
class PitchShiftOrTimeStretch(BatchRandTransform):
    "Randomly apply either `PitchShift` or `TimeStretch` to `TensorAudio` to minimize distortion"
    order, split_idx = 25, 0
    def __init__(self,
        p:float=0.4, # Per-item probability
        semitones:tuple[float,float]=(-4.0,4.0), # Random pitch shift range in semitones to compute efficient shifts
        bins_per_octave:int=12, # Number of steps per octave
        stretch_rates:tuple[float,float]=(0.5,2.0), # Random time stretch range to compute efficient stretches. Defaults to 50%-200% speed
        padmode:AudioPadMode=AudioPadMode.Repeat, # How to pad if necessary
        constant:Numeric=0, # Value for `AudioPadMode.Constant`
        split:int|None=None # Apply transform to `split` items at a time. Use to prevent OOM.
    ):
        super().__init__(p=p)
        if padmode not in [vol for vol in AudioPadMode]:
            raise ValueError(f"{padmode=} is not valid")
        self.bernoulli = Bernoulli(.5)
        self.shift   = PitchShift(1., semitones, bins_per_octave, padmode, constant, split)
        self.stretch = TimeStretch(1., stretch_rates, padmode, constant, split)

    def encodes(self, x:TensorAudio) -> Tensor:
        self.shift_smpl = self.bernoulli.sample((x.shape[0],))
        self.stretchs_smpl = (self.shift_smpl - 1) *-1
        self.shift_idxs, self.stretch_idxs = self.shift_smpl.bool(), self.stretchs_smpl.bool()
        if self.shift_smpl.sum() > 0:
            x[self.shift_idxs] = self.shift(x[self.shift_idxs], split_idx=0)
        if self.stretchs_smpl.sum() > 0:
            x[self.stretch_idxs] = self.stretch(x[self.stretch_idxs], split_idx=0)
        return x

# %% ../../nbs/audio.03_augment.ipynb 62
class TimeMasking(BatchRandTransform):
    "Randomly apply masking to a `TensorSpec` or `TensorMelSpec` in the time domain"
    split_idx, order = 0, 81
    def __init__(self,
        p:float=0.25, # Per-item probability
        max_mask:float=0.2, # Maximum possible length of the mask [0, max_mask)
        iid_masks:bool=True, # Apply different masks to each example/channel in the batch
        mask_value:int|None=0 # If None, random value between batch min and max
    ):
        store_attr(but='p')
        super().__init__(p=p)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        mask_param = int(self.max_mask * x.shape[-1])
        mv = random.randint(int(x.min), int(x.max)) if self.mask_value is None else self.mask_value
        if self.iid_masks and x.dim() == 4:
            return TAF.mask_along_axis_iid(x, mask_param, mv, 3)
        else:
            return TAF.mask_along_axis(x, mask_param, mv, 2)

# %% ../../nbs/audio.03_augment.ipynb 65
class FrequencyMasking(BatchRandTransform):
    "Randpmly apply masking to a `TensorSpec` or `TensorMelSpec` in the frequency domain"
    split_idx, order = 0, 82
    def __init__(self,
        p:float=0.25, # Per-item probability
        max_mask:float=0.2, # Maximum possible length of the mask [0, max_mask)
        iid_masks:bool=True, # Apply different masks to each example/channel in the batch
        mask_value:int|None=0 # If None, random value between batch min and max
    ):
        store_attr(but='p')
        super().__init__(p=p)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        mask_param = int(self.max_mask * x.shape[-2])
        mv = random.randint(int(x.min), int(x.max)) if self.mask_value is None else self.mask_value
        if self.iid_masks and x.dim() == 4:
            return TAF.mask_along_axis_iid(x, mask_param, mv, 2)
        else:
            return TAF.mask_along_axis(x, mask_param, mv, 1)

# %% ../../nbs/audio.03_augment.ipynb 69
class AmplitudeToDBMode(Enum):
    "All AmplitudeToDB modes as attributes to get tab-completion and typo-proofing",
    Power = 'power'
    Magnitude = 'magnitude'

# %% ../../nbs/audio.03_augment.ipynb 70
class AmplitudeToDB(DisplayedTransform):
    "Turn a `TensorSpec` or `TensorMelSpec` from the power/amplitude scale to the decibel scale"
    order = 98
    def __init__(self,
        top_db:float|None=None, # Minimum negative cut-off in decibels. A reasonable number is 80.
        mode:AmplitudeToDBMode=AmplitudeToDBMode.Power # Power is elementwise square of magnitude
    ):
        self.amdb = tatfms.AmplitudeToDB(mode.value, top_db)

    def encodes(self, x:TensorSpec|TensorMelSpec) -> Tensor:
        return self.amdb(x)

    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        self.amdb.to(device)

# %% ../../nbs/audio.03_augment.ipynb 74
class AudioNormalize(DisplayedTransform):
    "Normalize/denorm batch of `TensorSpec` or `TensorMelSpec`."
    parameters,order = L('mean', 'std'),99
    def __init__(self, mean=None, std=None, n_spec=1, axes=(0,2,3)):
        store_attr()
        self.c = 0

    @classmethod
    def from_stats(cls, mean, std, n_spec=1, dim=1, ndim=4, cuda=True):
        return cls(*broadcast_vec(dim, ndim, mean, std, cuda=cuda), n_spec=n_spec)

    def setups(self, dl:DataLoader):
        if self.mean is None or self.std is None:
            x,*_ = dl.one_batch()
            self.mean,self.std = x.mean(self.axes, keepdim=True),x.std(self.axes, keepdim=True)+1e-7

    def encodes(self, x:TensorSpec|TensorMelSpec):
        if self.n_spec > 1:
            i = self.c*x.channels
            j = (self.c+1)*x.channels
            self.c = self.c + 1 if self.c < self.n_spec-1 else 0
            return (x-self.mean[:,i:j,...]) / self.std[:,i:j,...]
        else:
            return (x-self.mean) / self.std

    # ToDo: figure out how to handle decodes
    # def decodes(self, x:TensorSpec|TensorMelSpec):
    #     f = to_cpu if x.device.type=='cpu' else noop
    #     return (x*f(self.std) + f(self.mean))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/callback.tracker.ipynb.

# %% ../../nbs/callback.tracker.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/callback.tracker.ipynb 3
from __future__ import annotations

from fastai.callback.core import Callback, CancelFitException
from fastai.callback.tracker import SaveModelCallback, TrackerCallback

from ..imports import *

# %% auto 0
__all__ = ['TerminateOnTrainNaN', 'SaveModelAtEnd', 'LastMetricCallback']

# %% ../../nbs/callback.tracker.ipynb 6
class TerminateOnTrainNaN(Callback):
    "A `Callback` that terminates training if the training loss is NaN and ignores valid loss."
    order, run_valid = -9, False
    def after_batch(self):
        "Test if `last_loss` is NaN and interrupts training."
        if torch.isinf(self.loss) or torch.isnan(self.loss): raise CancelFitException

# %% ../../nbs/callback.tracker.ipynb 9
class SaveModelAtEnd(SaveModelCallback):
    "A `SaveModelCallback` which only saves the model at the end so loggers can find it."
    order = TrackerCallback.order+1
    def __init__(self,
        fname='model', # Model filename
        with_opt=False # Include optimizer state
    ):
        # keep track of file path for loggers
        self.last_saved_path = None
        store_attr()

    def before_fit(self):
        pass

    def after_epoch(self):
        pass

    def after_fit(self, **kwargs):
        self.last_saved_path = self.learn.save(f'{self.fname}', with_opt=self.with_opt)

# %% ../../nbs/callback.tracker.ipynb 11
class LastMetricCallback(Callback):
    "A `Callback` which stores metrics by name in a `Learner.lastmetric` dictionary"
    order,remove_on_fetch,_only_train_loop = 60,True,True
    def __init__(self, metrics:Listified[str]='valid_loss'):
        self.metrics=L(metrics)

    def before_fit(self):
        "Prepare the monitored value(s)"
        self.run = not hasattr(self, "lr_finder") and not hasattr(self, "gather_preds")
        self.idx, self.learn.lastmetric = [], {}
        for m in self.metrics:
            assert m in self.recorder.metric_names[1:], f'Metric {m} does not exist'
            self.idx.append(list(self.recorder.metric_names[1:]).index(m))

    def after_fit(self):
        "Store the last the monitored value(s)"
        for i, idx in enumerate(self.idx):
            self.learn.lastmetric[self.metrics[i]] = self.recorder.values[-1][idx]
        self.run = True

    def after_fit_exception(self):
        try:
            self.after_fit()
        finally:
            self.run = True



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/callback.simpleprofiler.ipynb.

# %% auto 0
__all__ = ['SimpleProfilerPostCallback', 'SimpleProfilerCallback']

# %% ../../nbs/callback.simpleprofiler.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/callback.simpleprofiler.ipynb 3
import locale
import time
import pandas as pd
import numpy as np
from pathlib import Path
from packaging.version import parse

from fastcore.foundation import docs
from fastcore.basics import mk_class, noop, in_notebook

import fastai
from fastai.learner import Learner, Recorder
from fastai.callback.core import *

from ..imports import *
from ..utils import scale_time

if in_notebook():
    from IPython.display import display

# %% ../../nbs/callback.simpleprofiler.ipynb 9
if parse(fastai.__version__) >= parse('2.7.0'):
    _inner_loop = "before_draw before_batch after_pred after_loss before_backward after_cancel_backward after_backward before_step after_step after_cancel_batch after_batch".split()
else:
    _inner_loop = "before_draw before_batch after_pred after_loss before_backward before_step after_step after_cancel_batch after_batch".split()

# %% ../../nbs/callback.simpleprofiler.ipynb 10
if parse(fastai.__version__) >= parse('2.7.0'):
    _events = L.split('after_create before_fit before_epoch before_train before_draw before_batch after_pred after_loss \
        before_backward after_cancel_backward after_backward before_step after_cancel_step after_step \
        after_cancel_batch after_batch after_cancel_train after_train before_validate after_cancel_validate \
        after_validate after_cancel_epoch after_epoch after_cancel_fit after_fit')
else:
    _events = L.split('after_create before_fit before_epoch before_train before_draw before_batch after_pred after_loss \
        before_backward before_step after_cancel_step after_step after_cancel_batch after_batch after_cancel_train \
        after_train before_validate after_cancel_validate after_validate after_cancel_epoch \
        after_epoch after_cancel_fit after_fit')

mk_class('event', **_events.map_dict(),
         doc="All possible events as attributes to get tab-completion and typo-proofing")

# %% ../../nbs/callback.simpleprofiler.ipynb 12
@patch
def __call__(self:Callback, event_name):
    "Call `self.{event_name}` if it's defined"
    _run = (event_name not in _inner_loop or (self.run_train and getattr(self, 'training', True)) or
            (self.run_valid and not getattr(self, 'training', False)))
    res = None
    if self.run and _run:
        try: res = getattr(self, event_name, noop)()
        except (CancelBatchException, CancelEpochException, CancelFitException, CancelStepException, CancelTrainException, CancelValidException): raise
        except Exception as e:
            e.args = [f'Exception occured in `{self.__class__.__name__}` when calling event `{event_name}`:\n\t{e.args[0]}']
            raise
    if event_name=='after_fit': self.run=True #Reset self.run to True at each end of fit
    return res

# %% ../../nbs/callback.simpleprofiler.ipynb 14
@patch
def _call_one(self:Learner, event_name):
    if not hasattr(event, event_name): raise Exception(f'missing {event_name}')
    for cb in self.cbs.sorted('order'): cb(event_name)

# %% ../../nbs/callback.simpleprofiler.ipynb 16
@patch
def all_batches(self:Learner):
    self.n_iter = len(self.dl)
    if hasattr(self, 'simple_profiler'):
        self.it = iter(self.dl)
        for i in range(self.n_iter):
            self("before_draw")
            self.one_batch(i, next(self.it))
        del(self.it)
    else:
        for o in enumerate(self.dl): self.one_batch(*o)

# %% ../../nbs/callback.simpleprofiler.ipynb 18
_loop = ['Start Fit', 'before_fit', 'Start Epoch Loop', 'before_epoch', 'Start Train', 'before_train',
         'Start Batch Loop', 'before_draw', 'before_batch', 'after_pred', 'after_loss', 'before_backward',
         'before_step', 'after_step', 'after_cancel_batch', 'after_batch','End Batch Loop', 'End Train',
         'after_cancel_train', 'after_train', 'Start Valid', 'before_validate', 'Start Batch Loop',
         '**CBs same as train batch**', 'End Batch Loop', 'End Valid', 'after_cancel_validate',
         'after_validate', 'End Epoch Loop', 'after_cancel_epoch', 'after_epoch', 'End Fit',
         'after_cancel_fit', 'after_fit']

# %% ../../nbs/callback.simpleprofiler.ipynb 19
@patch
def show_training_loop(self:Learner):
    indent = 0
    for s in _loop:
        if s.startswith('Start'): print(f'{" "*indent}{s}'); indent += 2
        elif s.startswith('End'): indent -= 2; print(f'{" "*indent}{s}')
        else: print(f'{" "*indent} - {s:15}:', self.ordered_cbs(s))

# %% ../../nbs/callback.simpleprofiler.ipynb 21
_phase = ['fit', 'epoch', 'train', 'valid']
_epoch = ['train', 'valid']
_train = ['draw', 'batch', 'forward', 'loss', 'backward', 'opt_step', 'zero_grad']
_valid = ['draw', 'batch', 'predict', 'loss']

# %% ../../nbs/callback.simpleprofiler.ipynb 22
class SimpleProfilerPostCallback(Callback):
    "Pair with `SimpleProfilerCallback` to profile training performance. Removes itself after training is over."
    order,remove_on_fetch = Recorder.order-1,True
    def __init__(self, samples_per_second=True):
        store_attr()
        self._phase,self._train,self._valid = _phase,_train,_valid

    def before_fit(self):
        self.profiler = self.learn.simple_profiler
        self.has_logger = self.profiler.has_logger
        self.n_train_batches = len(self.dls.train)
        self.n_valid_batches = len(self.dls.valid)

    def after_train(self):
        self.profiler._raw_values['train'].append(time.perf_counter() - self.profiler._train_start)

    def after_validate(self):
        self.profiler._raw_values['valid'].append(time.perf_counter() - self.profiler._validate_start)

    def after_pred(self):
        if self.training:
            self.profiler._raw_values['train_forward'].append(time.perf_counter() - self.profiler._train_batch_start)
        else:
            self.profiler._raw_values['valid_predict'].append(time.perf_counter() - self.profiler._valid_batch_start)

        if self.training:
            self.profiler._train_loss_start = time.perf_counter()
        else:
            self.profiler._valid_loss_start = time.perf_counter()

    def after_loss(self):
        if self.training:
            self.profiler._raw_values['train_loss'].append(time.perf_counter() - self.profiler._train_loss_start)
        else:
            self.profiler._raw_values['valid_loss'].append(time.perf_counter() - self.profiler._valid_loss_start)

    def after_step(self):
        self.profiler._raw_values['train_opt_step'].append(time.perf_counter() - self.profiler._step_start)
        self.profiler._zero_start = time.perf_counter()

    def after_batch(self):
        if self.training:
            self.profiler._raw_values['train_batch'].append(time.perf_counter() - self.profiler._train_batch_start)
            if self.samples_per_second:
                self.profiler._raw_values['train_bs'].append(find_bs(self.learn.xb))
                if self.has_logger:
                    self.profiler._log_after_batch()
        else:
            self.profiler._raw_values['valid_batch'].append(time.perf_counter() - self.profiler._valid_batch_start)
            if self.samples_per_second:
                self.profiler._raw_values['valid_bs'].append(find_bs(self.learn.xb))

    def after_epoch(self):
        self.profiler._raw_values['epoch'].append(time.perf_counter() - self.profiler._epoch_start)

    def after_fit(self):
        self.profiler._raw_values['fit'].append(time.perf_counter() - self.profiler._fit_start)
        self.profiler._generate_report()
        if self.has_logger: self.profiler._log_after_fit()
        if not hasattr(self.learn, 'lr_finder'):
            self.profiler._display_report()
            self.learn.remove_cbs([SimpleProfilerCallback, SimpleProfilerPostCallback])

# %% ../../nbs/callback.simpleprofiler.ipynb 23
class SimpleProfilerCallback(Callback):
    """
    Adds a simple profiler to the fastai `Learner`. Optionally showing formatted report or saving unformatted results as csv.

    Pair with SimpleProfilerPostCallback to profile training performance.

    Post fit, access report & results via `Learner.simple_profile_report` & `Learner.simple_profile_results`.
    """
    order,remove_on_fetch = TrainEvalCallback.order+1,True
    def __init__(self,
        show_report=True, # Display formatted report post profile
        plain=False, # For Jupyter Notebooks, display plain report
        markdown=False, # Display markdown formatted report
        save_csv=False,  # Save raw results to csv
        csv_name='simple_profile.csv', # CSV save location
        logger_callback='wandb' # Log report and samples/second to `logger_callback` using `Callback.name`
    ):
        store_attr()
        self.csv_name = Path(csv_name)
        self._phase,self._train,self._valid = _phase,_train,_valid
        self._log_after_batch = getattr(self, f'_{self.logger_callback}_log_after_batch', noop)
        self._log_after_fit   = getattr(self, f'_{self.logger_callback}_log_after_fit', noop)

    def before_fit(self):
        self.has_logger = hasattr(self.learn, self.logger_callback) and not hasattr(self.learn, 'lr_finder') and not hasattr(self, "gather_preds")
        self._raw_values = dict()
        for p in _phase:
            self._raw_values[p] = []
        for p in _epoch:
            for a in getattr(self, f'_{p}'):
                self._raw_values[f'{p}_{a}'] = []
            self._raw_values[f'{p}_bs'] = []

        self._fit_start = time.perf_counter()

    def before_epoch(self):
        self._epoch_start = time.perf_counter()

    def before_train(self):
        self._train_start = time.perf_counter()

    def before_validate(self):
        self._validate_start = time.perf_counter()

    def before_draw(self):
        if self.training:
            self._train_draw_start = time.perf_counter()
        else:
            self._valid_draw_start = time.perf_counter()

    def before_batch(self):
        if self.training:
            self._raw_values['train_draw'].append(time.perf_counter() - self._train_draw_start)
        else:
            self._raw_values['valid_draw'].append(time.perf_counter() - self._valid_draw_start)

        if self.training:
            self._train_batch_start = time.perf_counter()
        else:
            self._valid_batch_start = time.perf_counter()

    def before_backward(self):
        self._backward_start = time.perf_counter()

    def before_step(self):
        self._raw_values['train_backward'].append(time.perf_counter() - self._backward_start)
        self._step_start = time.perf_counter()

    def after_batch(self):
        if self.training: self._raw_values['train_zero_grad'].append(time.perf_counter() - self._zero_start)

    def _train_samples_per_second(self, action):
        if action =='draw':
            bs = self._raw_values['train_bs'][-1]
            batch = self._raw_values['train_batch'][-1]
            return bs/(batch+self._raw_values[f'train_draw'][-1]) - bs/batch
        else:
            return self._raw_values[f'train_bs'][-1]/self._raw_values[f'train_{action}'][-1]

    def _generate_report(self):
        total_time = self._raw_values['fit'][0]
        self.report = pd.DataFrame(columns=['Phase', 'Action', 'Step', 'Mean Duration', 'Duration Std Dev',
                                            'Number of Calls', 'Samples/Second', 'Total Time', 'Percent of Total'])

        for p in _phase:
            if p == 'fit':
                self._append_to_df(['fit', p, p, 0, 0, 1, '-', total_time, f'{self._calc_percent(total_time):.0%}'])
            else:
                if p in _epoch:
                    self._append_to_df(self._create_overview_row('fit', 'epoch', p, self._raw_values[p], np.array(self._raw_values[f'{p}_bs'])))
                else:
                    self._append_to_df(self._create_overview_row('fit', p, p, self._raw_values[p], None))

        for p in _epoch:
            bs = np.array(self._raw_values[f'{p}_bs'])
            for i, s in enumerate(getattr(self, f'_{p}')):
                if s in ['draw', 'batch']: a = s
                self._append_to_df(self._create_detail_row(p, a, s, self._raw_values[f'{p}_{s}'], bs if s in _train else None))

        self.learn.simple_profile_results = self.report.copy()
        for c in ['Mean Duration', 'Duration Std Dev', 'Total Time']:
            self.report[c] = self.report[c].apply(scale_time)
        self.report[['Phase', 'Action']] = self.report[['Phase', 'Action']].where(~self.report[['Phase', 'Action']].duplicated(), '')
        self.report['Phase']  = self.report['Phase'].where(~self.report['Phase'].duplicated(), '')
        self.report['Step']   = self.report['Step'].where(self.report['Step'] != self.report['Action']).fillna('')
        self.report['Action'] = self.report['Action'].where(self.report['Phase'] != self.report['Action']).fillna('')

        self.learn.simple_profile_report = self.report

    def _display_report(self):
        if self.show_report:
            if self.markdown: print(self.report.to_markdown(index=False))
            else:
                if in_notebook() and not self.plain:
                    with pd.option_context('display.max_rows', len(self.report.index)):
                        s = self.report.style.set_caption("Simple Profiler Results").hide_index()
                        display(s)
                else:
                    print('Simple Profiler Results')
                    print(self.report.to_string(index=False))

        if self.save_csv:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.learn.simple_profile_results.to_csv(self.path/self.csv_name, index=False)

    def _append_to_df(self, row):
        self.report.loc[len(self.report.index)] = row

    def _calc_percent(self, time):
        return time / self._raw_values['fit'][0]

    def _create_overview_row(self, phase, action, step, input, bs=None):
        if bs is not None:
            draw = np.array(self._raw_values[f'{step}_draw'])
            batch = np.array(self._raw_values[f'{step}_batch'])
            sam_per_sec = f'{int(np.around(np.mean(bs/(draw+batch)))):,d}'
        else:
            sam_per_sec = '-'
        return [phase, action, step, np.mean(input), np.std(input), len(input), sam_per_sec,
                np.sum(input), f'{self._calc_percent(np.sum(input)):.0%}']

    def _create_detail_row(self, phase, action, step, input, bs=None):
        if bs is None or step=='zero_grad': sam_per_sec = '-'
        elif action=='draw':
            batch = np.array(self._raw_values[f'{phase}_batch'])
            sam_per_sec = f'{int(np.around(np.mean(bs/(np.array(input)+batch) - bs/batch))):,d}'
        else:
            sam_per_sec = f'{int(np.around(np.mean(bs/np.array(input)))):,d}'
        return [phase, action, step, np.mean(input), np.std(input), len(input), sam_per_sec,
                np.sum(input), f'{self._calc_percent(np.sum(input)):.0%}']

# %% ../../nbs/callback.simpleprofiler.ipynb 25
@patch
def profile(self:Learner,
        show_report=True, # Display formatted report post profile
        plain=False, # For Jupyter Notebooks, display plain report
        markdown=False, # Display markdown formatted report
        save_csv=False,  # Save raw results to csv
        csv_name='simple_profile.csv', # CSV save location
        samples_per_second=True, # Log samples/second for all actions & steps
        logger_callback='wandb' # Log report and samples/second to `logger_callback` using `Callback.name`
    ):
    "Run Simple Profiler when training. Simple Profiler removes itself when finished."
    self.add_cbs([SimpleProfilerCallback(show_report, plain, markdown, save_csv, csv_name, logger_callback),
                  SimpleProfilerPostCallback(samples_per_second)])
    return self

# %% ../../nbs/callback.simpleprofiler.ipynb 37
try:
    import wandb

    @patch
    def _wandb_log_after_batch(self:SimpleProfilerCallback):
        train_vals = {f'samples_per_second/train_{action}': self._train_samples_per_second(action) for action in _train[:-1]}
        wandb.log(train_vals, self.learn.wandb._wandb_step+1)

    @patch
    def _wandb_log_after_fit(self:SimpleProfilerCallback):
        report = wandb.Table(dataframe=self.learn.simple_profile_report)
        results = wandb.Table(dataframe=self.learn.simple_profile_results)

        wandb.log({"simple_profile_report": report})
        wandb.log({"simple_profile_results": results})
        wandb.log({}) # ensure sync
except:
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/optimizer.fused.ipynb.

# %% ../../nbs/optimizer.fused.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/optimizer.fused.ipynb 4
from __future__ import annotations
from typing import Optional, Dict

from fastcore.basics import partialler

from fastai.optimizer import (Optimizer, weight_decay, l2_reg, average_grad, sgd_step, momentum_step,
                              average_sqr_grad, rms_prop_step, step_stat, adam_step, radam_step,
                              qhadam_step, larc_layer_lr, larc_step, lamb_step, Lookahead)

from .torchscript import (JitOptimizer, radam_jit_step, sgd_jit_step, rmsprop_jit_step,
                                             adam_jit_step, radam_jit_step, qhadam_jit_step, larc_jit_step,
                                             lamb_jit_step, JitLookahead, ranger_jit_step)

from .foreach import (SGDForEachOptimizer, sgd_foreach_step, AdamForEachOptimizer,
                                         adam_foreach_step, RAdamForEachOptimizer, radam_foreach_step,
                                         LambForEachOptimizer, lamb_foreach_step, RangerForEachOptimizer,
                                         ranger_foreach_step)

from ..imports import *

# %% auto 0
__all__ = ['SGD', 'sgd', 'RMSProp', 'rmsprop', 'Adam', 'adam', 'RAdam', 'radam', 'QHAdam', 'qhadam', 'Larc', 'larc', 'Lamb',
           'lamb', 'Ranger', 'ranger']

# %% ../../nbs/optimizer.fused.ipynb 9
def SGD(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0., # Gradient moving average (β1) coefficient
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (SGDW) or L2 regularization (SGD)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|SGDForEachOptimizer|JitOptimizer:
    "A fastai SGD/SGDW optimizer with fused ForEach and TorchScript implementations"
    if foreach:
        return SGDForEachOptimizer(params, sgd_foreach_step, lr=lr, mom=mom, wd=wd, decouple_wd=decouple_wd)
    elif jit:
        return JitOptimizer(params, sgd_jit_step, lr=lr, mom=mom, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        if mom != 0: cbs.append(average_grad)
        cbs.append(sgd_step if mom==0 else momentum_step)
        return Optimizer(params, cbs, lr=lr, mom=mom, wd=wd)

# %% ../../nbs/optimizer.fused.ipynb 10
def sgd(
    mom:float=0., # Gradient moving average (β1) coefficient
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (SGDW) or L2 regularization (SGD)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|SGDForEachOptimizer|JitOptimizer:
    "Partial function for the SGD/SGDW optimizer with fused ForEach and TorchScript implementations"
    return partialler(SGD, mom=mom, wd=wd, decouple_wd=decouple_wd, jit=jit, foreach=foreach)

# %% ../../nbs/optimizer.fused.ipynb 13
def RMSProp(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0., # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "A fastai RMSProp/RMSPropW optimizer with a fused TorchScript implementation"
    if jit:
        return JitOptimizer(params, rmsprop_jit_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                            eps=eps, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        cbs += ([average_sqr_grad] if mom==0. else [average_grad, average_sqr_grad])
        cbs.append(rms_prop_step)
        return Optimizer(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, wd=wd, eps=eps)

# %% ../../nbs/optimizer.fused.ipynb 14
def rmsprop(
    mom:float=0., # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "Partial function for the RMSProp/RMSPropW optimizer with a fused TorchScript implementation"
    return partialler(RMSProp, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, decouple_wd=decouple_wd, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 17
def Adam(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0.01, # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (AdamW) or L2 regularization (Adam)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|AdamForEachOptimizer|JitOptimizer:
    "A fastai Adam/AdamW optimizer with fused ForEach and TorchScript implementations"
    if foreach:
        return AdamForEachOptimizer(params, adam_foreach_step, lr=lr, mom=mom,
                                    sqr_mom=sqr_mom, eps=eps, wd=wd, decouple_wd=decouple_wd)
    elif jit:
        return JitOptimizer(params, adam_jit_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                            eps=eps, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        cbs += [partial(average_grad, dampening=True), average_sqr_grad, step_stat, adam_step]
        return Optimizer(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd)

# %% ../../nbs/optimizer.fused.ipynb 18
def adam(
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0.01, # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|AdamForEachOptimizer|JitOptimizer:
    "Partial function for the Adam/AdamW optimizer with fused ForEach and TorchScript implementations"
    return partialler(Adam, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd,
                      decouple_wd=decouple_wd, foreach=foreach, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 21
def RAdam(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    beta:float=0., # Set to enable SAdam with native fastai RAdam
    decouple_wd:bool=True, # Apply true weight decay (RAdamW) or L2 regularization (RAdam)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|RAdamForEachOptimizer|JitOptimizer:
    "A fastai RAdam/RAdamW optimizer with fused ForEach and TorchScript implementations"
    if (foreach or jit) and beta != 0:
        raise ValueError(f'ForEach and TorchScript RAdam does not use {beta=}, set `jit` & `foreach` to False if beta!=0')
    if foreach:
        return RAdamForEachOptimizer(params, radam_foreach_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                                     eps=eps, wd=wd, decouple_wd=decouple_wd)
    if jit:
        return JitOptimizer(params, radam_jit_step, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps,
                            wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        cbs += [partial(average_grad, dampening=True), average_sqr_grad, step_stat, radam_step]
        return Optimizer(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, beta=beta)

# %% ../../nbs/optimizer.fused.ipynb 22
def radam(
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    beta:float=0., # Set to enable SAdam with native fastai RAdam
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|RAdamForEachOptimizer|JitOptimizer:
    "Partial function for the RAdam/RAdamW optimizer with fused ForEach and TorchScript implementations"
    return partialler(RAdam, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, beta=beta,
                      decouple_wd=decouple_wd, foreach=foreach, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 25
def QHAdam(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.999, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.999, # Gradient squared moving average (β2) coefficient
    nu_1:float=0.7, # QH immediate discount factor
    nu_2:float=1.0, # QH momentum discount factor
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (QHAdamW) or L2 regularization (QHAdam)
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "A fastai QHAdam/QHAdamW optimizer with a fused TorchScript implementation"
    if jit:
        return JitOptimizer(params, qhadam_jit_step, lr=lr, nu_1=nu_1, nu_2=nu_2, mom=mom,
                            sqr_mom=sqr_mom, eps=eps, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        cbs += [partial(average_grad, dampening=True), average_sqr_grad, step_stat, qhadam_step]
        return Optimizer(params, cbs, lr=lr, nu_1=nu_1, nu_2=nu_2, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd)

# %% ../../nbs/optimizer.fused.ipynb 26
def qhadam(
    mom:float=0.999, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.999, # Gradient squared moving average (β2) coefficient
    nu_1:float=0.7, # QH immediate discount factor
    nu_2:float=1.0, # QH momentum discount factor
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "Partial function for the QHAdam/QHAdamW optimizer with a fused TorchScript implementation"
    return partialler(QHAdam, mom=mom, sqr_mom=sqr_mom, nu_1=nu_1, nu_2=nu_2, eps=eps,
                      wd=wd, decouple_wd=decouple_wd, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 29
def Larc(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.9, # Gradient moving average (β1) coefficient
    clip:bool=True, # LARC if clip=True, LARS if clip=False
    trust_coeff:float=0.02, # Trust coeffiecnet for calculating layerwise LR
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay or L2 regularization
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "A fastai LARC/LARS optimizer with a fused TorchScript implementation"
    if jit:
        cb = partial(larc_jit_step, clip=clip)
        return JitOptimizer(params, cb, lr=lr, mom=mom, trust_coeff=trust_coeff,
                            eps=eps, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        if mom!=0.: cbs.append(average_grad)
        cbs += [partial(larc_layer_lr, clip=clip), larc_step]
        return Optimizer(params, cbs, lr=lr, mom=mom, trust_coeff=trust_coeff, eps=eps, wd=wd)

# %% ../../nbs/optimizer.fused.ipynb 30
def larc(
    mom:float=0.9, # Gradient moving average (β1) coefficient
    clip:bool=True, # LARC if clip=True, LARS if clip=False
    trust_coeff:float=0.02, # Trust coeffiecnet for calculating layerwise LR
    eps:float=1e-8, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|JitOptimizer:
    "Partial function for the LARC/LARS optimizer with a fused TorchScript implementation"
    return partialler(Larc, mom=mom, clip=clip, eps=eps, trust_coeff=trust_coeff,
                      wd=wd, decouple_wd=decouple_wd, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 33
def Lamb(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay or L2 regularization
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|LambForEachOptimizer|JitOptimizer:
    "A fastai LAMB optimizer with fused ForEach and TorchScript implementations"
    if foreach:
        return LambForEachOptimizer(params, lamb_foreach_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                                    eps=eps, wd=wd, decouple_wd=decouple_wd)
    if jit:
        return JitOptimizer(params, lamb_jit_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                            eps=eps, wd=wd, decouple_wd=decouple_wd)
    else:
        cbs = [weight_decay] if decouple_wd else [l2_reg]
        cbs += [partial(average_grad, dampening=True), average_sqr_grad, step_stat, lamb_step]
        return Optimizer(params, cbs, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd)

# %% ../../nbs/optimizer.fused.ipynb 34
def lamb(
    mom:float=0.9, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-5, # Added for numerical stability
    wd:float=0., # Optional weight decay (true or L2)
    decouple_wd:bool=True, # Apply true weight decay (RMSPropW) or L2 regularization (RMSProp)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Optimizer|LambForEachOptimizer|JitOptimizer:
    "Partial function for the LAMB optimizer with fused ForEach and TorchScript implementations"
    return partialler(Lamb, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd,
                      decouple_wd=decouple_wd, foreach=foreach, jit=jit)

# %% ../../nbs/optimizer.fused.ipynb 37
def Ranger(
    params:Listified[Tensor], # Model parameters or parameter groups
    lr:float, # Default learning rate
    mom:float=0.95, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-6, # Added for numerical stability
    wd:float=0.01, # Optional weight decay (true or L2)
    k:int=6, # How often to conduct Lookahead step
    alpha:float=0.5, # Slow weight moving average coefficient
    decouple_wd:bool=True, # Apply true weight decay (RAdamW) or L2 regularization (RAdam)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Lookahead|RangerForEachOptimizer|JitLookahead:
    "Convenience method for `Lookahead` with `RAdam` fused ForEach and TorchScript implementations"
    if foreach:
        return RangerForEachOptimizer(params, ranger_foreach_step, lr=lr, mom=mom, sqr_mom=sqr_mom,
                                      eps=eps, wd=wd, decouple_wd=decouple_wd, k=k, alpha=alpha)
    elif jit:
        return JitLookahead(params, ranger_jit_step, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps,
                            wd=wd, k=k, alpha=alpha, decouple_wd=decouple_wd)
    else:
        return Lookahead(RAdam(params, lr=lr, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd,
                               decouple_wd=decouple_wd),
                         k=k, alpha=alpha)

# %% ../../nbs/optimizer.fused.ipynb 38
def ranger(
    mom:float=0.95, # Gradient moving average (β1) coefficient
    sqr_mom:float=0.99, # Gradient squared moving average (β2) coefficient
    eps:float=1e-6, # Added for numerical stability
    wd:float=0.01, # Optional weight decay (true or L2)
    k:int=6, # How often to conduct Lookahead step
    alpha:float=0.5, # Slow weight moving average coefficient
    decouple_wd:bool=True, # Apply true weight decay (RAdamW) or L2 regularization (RAdam)
    foreach:bool=False, # Use fused ForEach implementation
    jit:bool=False # Use fused TorchScript implementation
) -> Lookahead|RangerForEachOptimizer|JitLookahead:
    "Partial function of the onvenience method for `Lookahead` with `RAdam` fused ForEach and TorchScript implementations"
    return partialler(Ranger, mom=mom, sqr_mom=sqr_mom, eps=eps, wd=wd, k=k,
                      alpha=alpha, decouple_wd=decouple_wd, foreach=foreach, jit=jit)
